import threading

thread_class = threading.Thread
pv_form = 'time'


def setup(logger):
    ...

def caget(*args, **kwargs):
    raise NotImplementedError

def caput(*args, **kwargs):
    raise NotImplementedError

def get_pv(*args, **kwargs):
    raise NotImplementedError





import logging
import time
from copy import copy

from ophyd import scaler
from ophyd.utils import enum
from .conftest import using_fake_epics_pv
from .test_utils import assert_OD_equal_ignore_ts


ScalerMode = enum(ONE_SHOT=0, AUTO_COUNT=1)

logger = logging.getLogger(__name__)


REAL_SCALER = False
scalers = ['XF:23ID2-ES{Sclr:1}']


@using_fake_epics_pv
def test_temp_scaler():
    # TODO fix
    scaler.EpicsScaler(scalers[0], name='test')


@using_fake_epics_pv
def test_scaler_functionality():
    sca = scaler.EpicsScaler(scalers[0], name='scaler',
                             read_attrs=['channels'])
    # hack the fake PV to know the enums
    sca.wait_for_connection()
    if not REAL_SCALER:
        sca.count_mode._read_pv.enum_strs = ['OneShot', 'AutoCount']
        sca.count_mode.put('OneShot')
        # pin the fake PVs by setting them
        attr_map = {getattr(sca, n).name: n for n in sca.configuration_attrs}
        cnf = {attr_map[k]: v['value']
               for k, v in sca.read_configuration().items()}

        sca.configure(cnf)

    sca.preset_time.put(5.2)

    logger.info('Counting in One-Shot mode for %f s...',
                sca.preset_time.get())
    sca.count.put(1)
    logger.info('Sleeping...')
    time.sleep(3)
    logger.info('Done sleeping. Stopping counter...')
    sca.stop()

    logger.info('Set mode to AutoCount')
    sca.count_mode.put(ScalerMode.AUTO_COUNT)
    sca.count.put(1)
    logger.info('Begin auto-counting (aka "background counting")...')
    time.sleep(2)
    logger.info('Set mode to OneShot')
    sca.count_mode.put(ScalerMode.ONE_SHOT)
    time.sleep(1)
    logger.info('Stopping (aborting) auto-counting.')
    sca.stop()

    logger.info('read() all channels in one-shot mode...')
    vals = sca.read()
    logger.info(vals)
    assert 'scaler_channels_chan1' in vals

    sca.report
    sca.read()

    assert copy(sca).prefix == sca.prefix
    assert copy(sca).read_attrs == sca.read_attrs
    assert copy(sca).configuration_attrs == sca.configuration_attrs
    repr(sca)
    str(sca)

    sca.stage()
    old, new = sca.configure({})
    sca.unstage()

    assert_OD_equal_ignore_ts(old, new)

    sca.stage()
    old_preset_time = sca.preset_time.get()
    old, new = sca.configure({'preset_time': 7})
    sca.unstage()

    assert old.pop('scaler_preset_time')['value'] == old_preset_time
    assert new.pop('scaler_preset_time')['value'] == 7
    assert_OD_equal_ignore_ts(old, new)

    sca.hints == {'fields': [sca.channels.name]}


@using_fake_epics_pv
def test_signal_separate():
    sca = scaler.EpicsScaler(scalers[0], name='scaler',
                             read_attrs=['channels.chan1'])
    sca.wait_for_connection()
    data = sca.read()
    assert 'scaler_channels_chan1' in data
    assert 'scaler_channels_chan2' not in data


@using_fake_epics_pv
def smoke_test_scalerCH():
    sca = scaler.ScalerCH(scalers[0])
    sca.wait_for_connection()
    data = sca.read()
    assert 'scaler_channels_chan1' in data
    assert 'scaler_channels_chan2' not in data



import sys
import time
import random
import logging
import pytest
import threading
from functools import wraps
import weakref

import numpy as np
import numpy.testing

from ophyd import get_cl, set_cl

logger = logging.getLogger(__name__)

_FAKE_PV_LIST = []


class FakeEpicsPV(object):
    _connect_delay = (0.05, 0.1)
    _update_rate = 0.1
    fake_values = (0.1, 0.2, 0.3)
    _pv_idx = 0
    auto_monitor = True

    def __init__(self, pvname, form=None,
                 callback=None, connection_callback=None,
                 auto_monitor=True, enum_strs=None,
                 **kwargs):

        self.callbacks = dict()

        global _FAKE_PV_LIST
        _FAKE_PV_LIST.append(self)

        self._pvname = pvname
        self._connection_callback = connection_callback
        self._form = form
        self._auto_monitor = auto_monitor
        self._value = self.fake_values[0]
        self._connected = False
        self._running = True
        self.enum_strs = enum_strs
        FakeEpicsPV._pv_idx += 1
        self._idx = FakeEpicsPV._pv_idx

        self._update = True

        self._lock = threading.Lock()
        self._thread = threading.Thread(target=self._update_loop)
        self._thread.daemon = True
        self._thread.start()

        if callback:
            self.add_callback(callback)

    def __del__(self):
        self.clear_callbacks()
        self._running = False

        try:
            self._thread.join()
            self._thread = None
        except Exception:
            pass

    def get_timevars(self):
        pass

    def get_ctrlvars(self):
        pass

    @property
    def connected(self):
        return self._connected

    def wait_for_connection(self, timeout=None):
        if self._pvname in ('does_not_connect', ):
            return False

        while not self._connected:
            time.sleep(0.05)

        return True

    def _update_loop(self):
        time.sleep(random.uniform(*self._connect_delay))
        if self._connection_callback is not None:
            self._connection_callback(pvname=self._pvname, conn=True, pv=self)

        if self._pvname in ('does_not_connect', ):
            return

        last_value = None
        while self._running:
            with self._lock:
                if self._update:
                    self._value = random.choice(self.fake_values)

                if self._value != last_value:
                    sys.stdout.flush()
                    self.run_callbacks()
                    last_value = self._value

                time.sleep(self._update_rate)
            self._connected = True

            time.sleep(0.01)

    @property
    def lower_ctrl_limit(self):
        return min(self.fake_values)

    @property
    def upper_ctrl_limit(self):
        return max(self.fake_values)

    def run_callbacks(self):
        for index in sorted(list(self.callbacks.keys())):
            if not self._running:
                break
            self.run_callback(index)

    def run_callback(self, index):
        fcn = self.callbacks[index]()
        if fcn is None:
            self.remove_callback(index)
            return

        kwd = dict(pvname=self._pvname,
                   count=1,
                   nelm=1,
                   type=None,
                   typefull=None,
                   ftype=None,
                   access='rw',
                   chid=self._idx,
                   read_access=True,
                   write_access=True,
                   value=self.value,
                   )

        kwd['cb_info'] = (index, self)
        if hasattr(fcn, '__call__'):
            fcn(**kwd)

    def add_callback(self, callback=None, index=None, run_now=False,
                     with_ctrlvars=True):
        if hasattr(callback, '__call__'):
            if index is None:
                index = 1
                if len(self.callbacks) > 0:
                    index = 1 + max(self.callbacks.keys())
            try:
                self.callbacks[index] = weakref.WeakMethod(callback)
            except TypeError:
                self.callbacks[index] = weakref.ref(callback)

        if run_now:
            if self.connected:
                self.run_callback(index)
        return index

    def remove_callback(self, index=None):
        self.callbacks.pop(index, None)

    def clear_callbacks(self):
        self.callbacks.clear()

    @property
    def precision(self):
        return 0

    @property
    def units(self):
        return str(None)

    @property
    def timestamp(self):
        return time.time()

    @property
    def pvname(self):
        return self._pvname

    @property
    def value(self):
        return self._value

    def __repr__(self):
        return '<FakePV %s value=%s>' % (self._pvname, self.value)

    def get(self, as_string=False, use_numpy=False,
            use_monitor=False):
        if as_string:

            if isinstance(self.value, list):
                if self.enum_strs:
                    return [self.enum_strs[_] for _ in self.value]
                return list(self.value)
            if isinstance(self.value, str):
                return self.value
            else:
                if self.enum_strs:
                    return self.enum_strs[self.value]
                return str(self.value)
        elif use_numpy:
            return np.array(self.value)
        else:
            return self.value

    def put(self, value, wait=False, timeout=30.0,
            use_complete=False, callback=None, callback_data=None):

        with self._lock:
            self._update = False
            self._value = value


class FakeEpicsWaveform(FakeEpicsPV):
    strings = ['abcd', 'efgh', 'ijkl']
    fake_values = [[ord(c) for c in s] + [0]
                   for s in strings]
    auto_monitor = False
    form = 'time'


def _cleanup_fake_pvs():
    pvs = list(_FAKE_PV_LIST)
    del _FAKE_PV_LIST[:]

    for pv in pvs:
        pv.clear_callbacks()
        pv._running = False
        pv._connection_callback = None

    for pv in pvs:
        try:
            pv._thread.join()
            pv._thread = None
        except Exception:
            pass


def using_fake_epics_pv(fcn):
    @wraps(fcn)
    def wrapped(*args, **kwargs):
        cl = get_cl()
        get_pv_backup = cl.get_pv

        def _fake_get_pv(pvname, form='time', connect=False,
                         context=False, timout=5.0, **kw):
            return FakeEpicsPV(pvname, form=form, **kw)
        cl.get_pv = _fake_get_pv
        try:
            return fcn(*args, **kwargs)
        finally:
            cl.get_pv = get_pv_backup
            _cleanup_fake_pvs()

    return wrapped


def using_fake_epics_waveform(fcn):
    @wraps(fcn)
    def wrapped(*args, **kwargs):
        cl = get_cl()
        get_pv_backup = cl.get_pv

        def _fake_get_pv(pvname, form='time', connect=False,
                         context=False, timout=5.0, **kw):
            return FakeEpicsWaveform(pvname, form=form, **kw)

        cl.get_pv = _fake_get_pv
        try:
            return fcn(*args, **kwargs)
        finally:
            cl.get_pv = get_pv_backup
            _cleanup_fake_pvs()

    return wrapped


@pytest.fixture()
def hw():
    from ophyd.sim import hw
    return hw()


@pytest.fixture(params=['caproto', 'pyepics'], autouse=True)
def cl_selector(request):
    cl_name = request.param
    if cl_name == 'caproto':
        pytest.importorskip('caproto')
    elif cl_name == 'pyepics':
        pytest.importorskip('epics')
    set_cl(cl_name)
    yield
    set_cl()


class AssertTools:
    @staticmethod
    def assertEquals(a, b):
        assert a == b

    @staticmethod
    def assertEqual(a, b):
        assert a == b

    @staticmethod
    def assertNotEqual(a, b):
        assert a != b

    @staticmethod
    def assertRaises(Etype, func, *args, **kwargs):
        with pytest.raises(Etype):
            func(*args, **kwargs)

    @staticmethod
    def assertIn(val, target):
        assert val in target

    @staticmethod
    def assertIs(a, b):
        assert a is b

    @staticmethod
    def assertTrue(v):
        assert v

    @staticmethod
    def assertFalse(v):
        assert not v

    @staticmethod
    def assertGreater(a, b):
        assert a > b

    @staticmethod
    def assertAlmostEqual(a, b):
        numpy.testing.assert_almost_equal(a, b)

    @staticmethod
    def skipTest(msg):
        pytest.skip(msg)



import collections
import contextlib
import functools
import time as ttime
import logging
import textwrap
from enum import Enum
from collections import (OrderedDict, namedtuple)
import warnings

from .ophydobj import OphydObject, Kind
from .status import DeviceStatus, StatusBase
from .utils import (ExceptionBundle, set_and_wait, RedundantStaging,
                    doc_annotation_forwarder)

from typing import Dict, List, Any, TypeVar, Tuple
from types import SimpleNamespace
from collections.abc import MutableSequence
from itertools import groupby

A, B = TypeVar('A'), TypeVar('B')
ALL_COMPONENTS = object()


class OrderedDictType(Dict[A, B]):
    ...


logger = logging.getLogger(__name__)


class Staged(Enum):
    """Three-state switch"""
    yes = 'yes'
    no = 'no'
    partially = 'partially'


class Component:
    '''A descriptor representing a device component (or signal)

    Unrecognized keyword arguments will be passed directly to the component
    class initializer.

    Parameters
    ----------
    cls : class
        Class of signal to create.  The required signature of
        `cls.__init__` is (if `suffix` is given)::

            def __init__(self, pv_name, parent=None, **kwargs):

        or (if suffix is None) ::

            def __init__(self, parent=None, **kwargs):

        The class may have a `wait_for_connection()` which is called
        during the component instance creation.

    suffix : str, optional
        The PV suffix, which gets appended onto ``parent.prefix`` to
        generate the final PV that the instance component will bind to.
        Also see ``add_prefix``

    lazy : bool, optional
        Lazily instantiate the signal. If ``False``, the signal will be
        instantiated upon component instantiation

    trigger_value : any, optional
        Mark as a signal to be set on trigger. The value is sent to the signal
        at trigger time.

    add_prefix : sequence, optional
        Keys in the kwargs to prefix with the Device PV prefix during
        creation of the component instance.
        Defaults to ``('suffix', 'write_pv', )``

    doc : str, optional
        string to attach to component DvcClass.component.__doc__
    '''

    def __init__(self, cls, suffix=None, *, lazy=False, trigger_value=None,
                 add_prefix=None, doc=None, kind=Kind.normal, **kwargs):
        self.attr = None  # attr is set later by the device when known
        self.cls = cls
        self.kwargs = kwargs
        self.lazy = lazy
        self.suffix = suffix
        self.doc = doc
        self.trigger_value = trigger_value  # TODO discuss
        self.kind = _ensure_kind(kind)
        if add_prefix is None:
            add_prefix = ('suffix', 'write_pv')
        self.add_prefix = tuple(add_prefix)

    def maybe_add_prefix(self, instance, kw, suffix):
        """Add prefix to a suffix if kw is in self.add_prefix

        Parameters
        ----------
        instance : Device
            The instance to extract the prefix to maybe append to the
            suffix from.

        kw : str
            The key of associated with the suffix.  If this key is
            self.add_prefix than prepend the prefix to the suffix and
            return, else just return the suffix.

        suffix : str
            The suffix to maybe have something prepended to.

        Returns
        -------
        str
        """
        if kw in self.add_prefix:
            return '{prefix}{suffix}'.format(prefix=instance.prefix,
                                             suffix=suffix)
        return suffix

    def create_component(self, instance):
        '''Create a component for the instance'''
        kwargs = self.kwargs.copy()
        kwargs['name'] = '{}_{}'.format(instance.name, self.attr)
        kwargs['kind'] = instance._initial_state[self.attr].kind

        for kw, val in list(kwargs.items()):
            kwargs[kw] = self.maybe_add_prefix(instance, kw, val)

        if (isinstance(self.cls, type) and  # is a class
                issubclass(self.cls, DynamicDeviceComponent)):
            cpt_inst = self.cls(self.suffix).create_component(self)

        elif self.suffix is not None:
            pv_name = self.maybe_add_prefix(instance, 'suffix', self.suffix)
            cpt_inst = self.cls(pv_name, parent=instance, **kwargs)
        else:
            cpt_inst = self.cls(parent=instance, **kwargs)

        if self.lazy and hasattr(self.cls, 'wait_for_connection'):
            cpt_inst.wait_for_connection()

        return cpt_inst

    def make_docstring(self, parent_class):
        if self.doc is not None:
            return self.doc

        doc = ['{} attribute'.format(self.__class__.__name__),
               '::',
               '',
               ]

        doc.append(textwrap.indent(repr(self), prefix=' ' * 4))
        doc.append('')
        return '\n'.join(doc)

    def __repr__(self):

        kw_str = ', '.join('{}={!r}'.format(k, v) for k, v in self.kwargs.items() \
                           if k not in ['read_attrs','configuration_attrs'])

        if self.suffix is not None:
            suffix_str = '{!r}'.format(self.suffix)
            if self.kwargs:
                suffix_str += ', '
        else:
            suffix_str = ''

        if suffix_str or kw_str:
            arg_str = ', {}{}'.format(suffix_str, kw_str)
        else:
            arg_str = ''
        arg_str += ', kind={}'.format(self.kind)

        return ('{self.__class__.__name__}({self.cls.__name__}{arg_str})'
                ''.format(self=self, arg_str=arg_str))

    __str__ = __repr__

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if self.attr not in instance._signals:
            instance._signals[self.attr] = self.create_component(instance)

        return instance._signals[self.attr]

    def __set__(self, instance, owner):
        raise RuntimeError('Use .put()')


class FormattedComponent(Component):
    '''A Component which takes a dynamic format string

    This differs from Component in that the parent prefix is not automatically
    added onto the Component suffix. Additionally, `str.format()` style strings
    are accepted, allowing access to Device instance attributes:

    >>> from ophyd import (Component as C, FormattedComponent as FC)
    >>> class MyDevice(Device):
    ...     # A normal component, where 'suffix' is added to prefix verbatim
    ...     cpt = C(EpicsSignal, 'suffix')
    ...     # A formatted component, where 'self' refers to the Device instance
    ...     ch = FC(EpicsSignal, '{self.prefix}{self._ch_name}')
    ...
    ...     def __init__(self, prefix, ch_name=None, **kwargs):
    ...         self._ch_name = ch_name
    ...         super().__init__(prefix, **kwargs)

    >>> dev = MyDevice('prefix:', ch_name='some_channel', name='dev')
    >>> print(dev.cpt.pvname)
    prefix:suffix
    >>> print(dev.ch.pvname)
    prefix:some_channel

    For additional documentation, refer to Component.
    '''

    def maybe_add_prefix(self, instance, kw, suffix):
        if kw not in self.add_prefix:
            return suffix

        return suffix.format(self=instance)


class DynamicDeviceComponent:
    '''An Device component that dynamically creates a OphydDevice

    Parameters
    ----------
    defn : OrderedDict
        The definition of all attributes to be created, in the form of::

            defn['attribute_name'] = (SignalClass, pv_suffix, keyword_arg_dict)

        This will create an attribute on the sub-device of type `SignalClass`,
        with a suffix of pv_suffix, which looks something like this::

            parent.sub.attribute_name = Cpt(SignalClass, pv_suffix, **keyword_arg_dict)

        Keep in mind that this is actually done in the metaclass creation, and
        not exactly as written above.
    clsname : str, optional
        The name of the class to be generated
        This defaults to {parent_name}{this_attribute_name.capitalize()}
    doc : str, optional
        The docstring to put on the dynamically generated class
    default_read_attrs : list, optional
        A class attribute to put on the dynamically generated class
    default_configuration_attrs : list, optional
        A class attribute to put on the dynamically generated class
    '''

    def __init__(self, defn, *, clsname=None, doc=None, kind=Kind.normal,
                 default_read_attrs=None, default_configuration_attrs=None):
        self.defn = defn
        self.clsname = clsname
        self.attr = None  # attr is set later by the device when known
        self.lazy = False
        self.doc = doc
        if isinstance(default_read_attrs, collections.Iterable):
            default_read_attrs = tuple(default_read_attrs)
        if isinstance(default_configuration_attrs, collections.Iterable):
            default_configuration_attrs = tuple(default_configuration_attrs)
        self.default_read_attrs = default_read_attrs
        self.default_configuration_attrs = default_configuration_attrs
        self.kind = _ensure_kind(kind)

        # TODO: component compatibility
        self.trigger_value = None
        self.attrs = list(defn.keys())

    def make_docstring(self, parent_class):
        if self.doc is not None:
            return self.doc

        doc = ['{} comprised of'.format(self.__class__.__name__),
               '::',
               '',
               ]

        doc.append(textwrap.indent(repr(self), prefix=' ' * 4))
        doc.append('')
        return '\n'.join(doc)

    def __repr__(self):
        doc = []
        for attr, (cls, suffix, kwargs) in self.defn.items():
            kw_str = ', '.join('{}={!r}'.format(k, v)
                               for k, v in kwargs.items())
            if suffix is not None:
                suffix_str = '{!r}'.format(suffix)
                if kwargs:
                    suffix_str += ', '
            else:
                suffix_str = ''

            if suffix_str or kw_str:
                arg_str = ', {}{}'.format(suffix_str, kw_str)
            else:
                arg_str = ''

            doc.append('{attr} = Component({cls.__name__}{arg_str})'
                       ''.format(attr=attr, cls=cls, arg_str=arg_str))

        return '\n'.join(doc)

    def create_attr(self, attr_name):
        cls, suffix, kwargs = self.defn[attr_name]
        inst = Component(cls, suffix, **kwargs)
        inst.attr = attr_name
        return inst

    def create_component(self, instance):
        '''Create a component for the instance'''
        clsname = self.clsname
        if clsname is None:
            # make up a class name based on the instance's class name
            clsname = ''.join((instance.__class__.__name__,
                               self.attr.capitalize()))

            # TODO: and if the attribute has any underscores, convert that to
            #       camelcase

        docstring = self.doc
        if docstring is None:
            docstring = '{} sub-device'.format(clsname)

        clsdict = OrderedDict(
            __doc__=docstring,
            _default_read_attrs=self.default_read_attrs,
            _default_configuration_attrs=self.default_configuration_attrs
        )

        for attr in self.defn.keys():
            clsdict[attr] = self.create_attr(attr)

        cls = type(clsname, (Device, ), clsdict)
        return cls(instance.prefix,
                   name='{}_{}'.format(instance.name, self.attr),
                   parent=instance,
                   kind=instance._initial_state[self.attr].kind)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if self.attr not in instance._signals:
            instance._signals[self.attr] = self.create_component(instance)

        return instance._signals[self.attr]

    def __set__(self, instance, value):
        raise RuntimeError('Use .put()')


class ComponentMeta(type):
    '''Creates attributes for Components by inspecting class definition'''

    @classmethod
    def __prepare__(self, name, bases):
        '''Prepare allows the class attribute dictionary to be ordered as
        defined by the user'''
        return OrderedDict()

    def __new__(cls, name, bases, clsdict):
        clsobj = super().__new__(cls, name, bases, clsdict)

        # This attrs are defined at instanitation time and must not
        # collide with class attributes.
        INSTANCE_ATTRS = ['name', 'parent', 'component_names', '_signals',
                          '_sig_attrs',
                          '_sub_devices']
        # These attributes are part of the bluesky interface and cannot be
        # used as component names.
        RESERVED_ATTRS = ['read', 'describe', 'trigger',
                          'configure', 'read_configuration',
                          'describe_configuration', 'describe_collect',
                          'set', 'stage', 'unstage', 'pause', 'resume',
                          'kickoff', 'complete', 'collect', 'position', 'stop',
                          # from OphydObject
                          'subscribe', 'clear_sub', 'event_types', 'root',
                          # for back-compat
                          'signal_names']
        for attr in INSTANCE_ATTRS:
            if attr in clsdict:
                raise TypeError("The attribute name %r is reserved for "
                                "use by the Device class. Choose a different "
                                "name." % attr)

        clsobj._sig_attrs = OrderedDict()
        # this is so that the _sig_attrs class attribute includes the
        # sigattrs from all of it's class-inheritance-parents so we do
        # not have to do this look up everytime we look at it.
        for base in reversed(bases):
            if not hasattr(base, '_sig_attrs'):
                continue

            for attr, cpt in base._sig_attrs.items():
                clsobj._sig_attrs[attr] = cpt

        # map component classes to their attribute names from this class
        for attr, value in clsdict.items():
            if isinstance(value, (Component, DynamicDeviceComponent)):
                if attr in RESERVED_ATTRS:
                    raise TypeError("The attribute name %r is part of the "
                                    "bluesky interface and cannot be used as "
                                    "the name of a component. Choose a "
                                    "different name." % attr)
                clsobj._sig_attrs[attr] = value

        for cpt_attr, cpt in clsobj._sig_attrs.items():
            # Notify the component of their attribute name
            cpt.attr = cpt_attr

        # List Signal attribute names.
        clsobj.component_names = tuple(clsobj._sig_attrs.keys())

        # The namedtuple associated with the device
        clsobj._device_tuple = namedtuple(
            name + 'Tuple',
            [comp for comp in clsobj.component_names
             if not comp.startswith('_')])
        # Finally, create all the component docstrings
        for cpt in clsobj._sig_attrs.values():
            cpt.__doc__ = cpt.make_docstring(clsobj)

        # List the attributes that are Devices (not Signals).
        # This list is used by stage/unstage. Only Devices need to be staged.
        clsobj._sub_devices = []
        for attr, cpt in clsobj._sig_attrs.items():
            if (isinstance(cpt, Component) and
                    (not isinstance(cpt.cls, type) or  # not a class
                     not issubclass(cpt.cls, Device))):  # not a Device
                continue
            clsobj._sub_devices.append(attr)

        return clsobj


# These stub 'Interface' classes are the apex of the mro heirarchy for
# their respective methods. They make multiple interitance more
# forgiving, and let us define classes that customize these methods
# but are not full Devices.


class BlueskyInterface:
    """Classes that inherit from this can safely customize the
    these methods without breaking mro.

    """
    def __init__(self, *args, **kwargs):
        # Subclasses can populate this with (signal, value) pairs, to be
        # set by stage() and restored back by unstage().
        self.stage_sigs = OrderedDict()

        self._staged = Staged.no
        self._original_vals = OrderedDict()
        super().__init__(*args, **kwargs)

    def trigger(self) -> StatusBase:
        """Trigger the device and return status object

        This method is responsible for implementing 'trigger' or
        'acquire' functionality of this device.

        If there is an appreciable time between triggering the device
        and it being able to be read (via the
        :meth:`~BlueskyInterface.read` method) then this method is
        also responsible for arranging that the
        :obj:`~ophyd.status.StatusBase` object returned my this method
        is notified when the device is ready to be read.

        If there is no delay between triggering and being readable,
        then this method must return a :obj:`~ophyd.status.SatusBase`
        object which is already completed.

        Returns
        -------
        status : StatusBase
            :obj:`~ophyd.status.StatusBase` object which will be marked
            as complete when the device is ready to be read.

        """
        pass

    def read(self) -> OrderedDictType[str, Dict[str, Any]]:
        """Read data from the device

        This method is expected to be as instantaneous as possible,
        with any substantial acquisition time taken care of in
        :meth:`~BlueskyInterface.trigger`.

        The `OrderedDict` returned by this method must have identical
        keys (in the same order) as the `OrderedDict` returned by
        :meth:`~BlueskyInterface.describe()`.

        By convention, the first key in the return is the 'primary' key
        and maybe used by heuristics in :mod:`bluesky`.

        The values in the ordered dictionary must be dict (-likes) with the
        keys ``{'value', 'timestamp'}``.  The ``'value'`` may have any type,
        the timestamp must be a float UNIX epoch timestamp in UTC.

        Returns
        -------
        data : OrderedDict
            The keys must be strings and the values must be dict-like
            with the keys ``{'value', 'timestamp'}``

        """
        return OrderedDict()

    def describe(self) -> OrderedDictType[str, Dict[str, Any]]:
        """Provide schema and meta-data for :meth:`~BlueskyInterface.read`

        This keys in the `OrderedDict` this method returns must match the
        keys in the `OrderedDict` return by :meth:`~BlueskyInterface.read`.

        This provides schema related information, (ex shape, dtype), the
        source (ex PV name), and if available, units, limits, precision etc.

        Returns
        -------
        data_keys : OrderedDict
            The keys must be strings and the values must be dict-like
            with the ``event_model.event_descriptor.data_key`` schema.
        """
        return OrderedDict()

    def stage(self) -> List[object]:
        """Stage the device for data collection.

        This method is expected to put the device into a state where
        repeated calls to :meth:`~BlueskyInterface.trigger` and
        :meth:`~BlueskyInterface.read` will 'do the right thing'.

        Staging not idempotent and should raise
        :obj:`RedundantStaging` if staged twice without an
        intermediate :meth:`~BlueskyInterface.unstage`.

        This method should be as fast as is feasible as it does not return
        a status object.

        The return value of this is a list of all of the (sub) devices
        stage, including it's self.  This is used to ensure devices
        are not staged twice by the :obj:`~bluesky.run_engine.RunEngine`.

        This is an optional method, if the device does not need
        staging behavior it should not implement `stage` (or
        `unstage`).

        Returns
        -------
        devices : list
            list including self and all child devices staged

        """
        if self._staged == Staged.no:
            pass  # to short-circuit checking individual cases
        elif self._staged == Staged.yes:
            raise RedundantStaging("Device {!r} is already staged. "
                                   "Unstage it first.".format(self))
        elif self._staged == Staged.partially:
            raise RedundantStaging("Device {!r} has been partially staged. "
                                   "Maybe the most recent unstaging "
                                   "encountered an error before finishing. "
                                   "Try unstaging again.".format(self))
        self.log.debug("Staging %s", self.name)
        self._staged = Staged.partially

        # Resolve any stage_sigs keys given as strings: 'a.b' -> self.a.b
        stage_sigs = OrderedDict()
        for k, v in self.stage_sigs.items():
            if isinstance(k, str):
                # Device.__getattr__ handles nested attr lookup
                stage_sigs[getattr(self, k)] = v
            else:
                stage_sigs[k] = v

        # Read current values, to be restored by unstage()
        original_vals = {sig: sig.get() for sig in stage_sigs}

        # We will add signals and values from original_vals to
        # self._original_vals one at a time so that
        # we can undo our partial work in the event of an error.

        # Apply settings.
        devices_staged = []
        try:
            for sig, val in stage_sigs.items():
                self.log.debug("Setting %s to %r (original value: %r)",
                               self.name,
                               val, original_vals[sig])
                set_and_wait(sig, val)
                # It worked -- now add it to this list of sigs to unstage.
                self._original_vals[sig] = original_vals[sig]
            devices_staged.append(self)

            # Call stage() on child devices.
            for attr in self._sub_devices:
                device = getattr(self, attr)
                if hasattr(device, 'stage'):
                    device.stage()
                    devices_staged.append(device)
        except Exception:
            self.log.debug("An exception was raised while staging %s or "
                           "one of its children. Attempting to restore "
                           "original settings before re-raising the "
                           "exception.", self.name)
            self.unstage()
            raise
        else:
            self._staged = Staged.yes
        return devices_staged

    def unstage(self) -> List[object]:
        """Unstage the device.

        This method returns the device to the state it was prior to the
        last `stage` call.

        This method should be as fast as feasible as it does not
        return a status object.

        This method must be idempotent, multiple calls (without a new
        call to 'stage') have no effect.

        Returns
        -------
        devices : list
            list including self and all child devices unstaged

        """
        self.log.debug("Unstaging %s", self.name)
        self._staged = Staged.partially
        devices_unstaged = []

        # Call unstage() on child devices.
        for attr in self._sub_devices[::-1]:
            device = getattr(self, attr)
            if hasattr(device, 'unstage'):
                device.unstage()
                devices_unstaged.append(device)

        # Restore original values.
        for sig, val in reversed(list(self._original_vals.items())):
            self.log.debug("Setting %s back to its original value: %r)",
                           self.name,
                           val)
            set_and_wait(sig, val)
            self._original_vals.pop(sig)
        devices_unstaged.append(self)

        self._staged = Staged.no
        return devices_unstaged

    def pause(self) -> None:
        """Attempt to 'pause' the device.

        This is called when ever the
        :obj:`~bluesky.run_engine.RunEngine` is interrupted.

        A device may have internal state that means plans can not
        safely be re-wound.  This method may: put the device in a
        'paused' state and/or raise
        :obj:`~bluesky.run_engine.NoReplayAllowed` to indicate that
        the plan can not be rewound.

        Raises
        ------
        bluesky.run_engine.NoReplayAllowed

        """
        pass

    def resume(self) -> None:
        """Resume a device from a 'paused' state

        This is called by the :obj:`bluesky.run_engine.RunEngine`
        when it resumes from an interruption and is responsible for
        ensuring that the device is ready to take data again.
        """
        pass

    def _validate_kind(self, val):
        return super()._validate_kind(val)


class GenerateDatumInterface:
    """Classes that inherit from this can safely customize the
    `generate_datum` method without breaking mro. If used along with the
    BlueskyInterface, inherit from this second."""
    def generate_datum(self, key, timestamp, datum_kwargs):
        pass


class Device(BlueskyInterface, OphydObject, metaclass=ComponentMeta):
    """Base class for device objects

    This class provides attribute access to one or more Signals, which can be
    a mixture of read-only and writable. All must share the same base_name.

    Parameters
    ----------
    prefix : str, optional
        The PV prefix for all components of the device
    name : str, keyword only
        The name of the device
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.
    read_attrs : sequence of attribute names
        DEPRECATED
        the components to include in a normal reading (i.e., in ``read()``)
    configuration_attrs : sequence of attribute names
        DEPRECATED
        the components to be read less often (i.e., in
        ``read_configuration()``) and to adjust via ``configure()``
    parent : instance or None, optional
        The instance of the parent device, if applicable
    """

    SUB_ACQ_DONE = 'acq_done'  # requested acquire

    # over ride in sub-classes to control the default
    # contents of read and configuration attrs lists

    # To use the new 'kind' parameter, set these equal to the sentinel
    # REPSECT_KIND, defined in this module.

    # If `None`, defaults to `self.component_names'
    _default_read_attrs = None
    # If `None`, defaults to `[]`
    _default_configuration_attrs = None

    def __init__(self, prefix='', *, name, kind=None,
                 read_attrs=None, configuration_attrs=None,
                 parent=None, **kwargs):
        # Store EpicsSignal objects (only created once they are accessed)
        self._signals = {}
        self._initial_state = {k: SimpleNamespace(kind=cpt.kind)
                               for k, cpt in self._sig_attrs.items()}
        self.prefix = prefix
        if self.component_names and prefix is None:
            raise ValueError('Must specify prefix if device signals are being '
                             'used')

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)

        # The logic of these if blocks is:
        # - If the user did not provide read_attrs, fall back on the default
        # specified by the class, which ultimately falls back to Device, if no
        # subclass overrides it. Either way, we now have a read_attrs.
        # - If it is set to the sentinel ALL_COMPONENTS, ignore whatever the
        # 'kind' settings are; just include everything. This is an escape catch
        # for getting what _default_read_attrs=None used to do before 'kind'
        # was implemented.
        # - If it is set to a list, ignore whatever the 'kind' settings are and
        # just include that list.
        # - If it is set to None, respect whatever the 'kind' settings of the
        # components are.

        # If any sub-Devices are to be removed from configuration_attrs and
        # read_attrs, we have to remove them from read_attrs first, or they
        # will not allow themselves to be removed from configuration_attrs.
        if read_attrs is None:
            read_attrs = self._default_read_attrs
        if read_attrs is ALL_COMPONENTS:
            read_attrs = self.component_names
        if read_attrs is not None:
            self.read_attrs = list(read_attrs)

        if configuration_attrs is None:
            configuration_attrs = self._default_configuration_attrs
        if configuration_attrs is ALL_COMPONENTS:
            configuration_attrs = self.component_names
        if configuration_attrs is not None:
            self.configuration_attrs = list(configuration_attrs)

        # Instantiate non-lazy signals
        [getattr(self, attr) for attr, cpt in self._sig_attrs.items()
         if not cpt.lazy]

    def _validate_kind(self, val):
        if isinstance(val, str):
            val = getattr(Kind, val.lower())
        if Kind.normal & val:
            val = val | Kind.config
        return super()._validate_kind(val)

    @property
    def read_attrs(self):
        return self.OphydAttrList(self, Kind.normal, Kind.hinted,
                                  'read_attrs')

    @read_attrs.setter
    def read_attrs(self, val):
        self.__attr_list_helper(val, Kind.normal, Kind.hinted, 'read_attrs')

    @property
    def configuration_attrs(self):
        return self.OphydAttrList(self, Kind.config, Kind.config,
                                  'configuration_attrs')

    @configuration_attrs.setter
    def configuration_attrs(self, val):
        self.__attr_list_helper(val, Kind.config, Kind.config,
                                'configuration_attrs')

    def __attr_list_helper(self, val, set_kind, unset_kind, recurse_name):
        val = set(val)
        cn = set(self.component_names)

        for c in cn:
            if c in val:
                _lazy_get(self, c).kind |= set_kind
            else:
                _lazy_get(self, c).kind &= ~unset_kind

        # now look at everything else, presumably things with dots
        extra = val - cn
        fail = set(c for c in extra if '.' not in c)
        if fail:

            raise ValueError("You asked to add the components {fail} "
                             "to {recurse_name} "
                             "on {self.name!r}, but there is no such child in "
                             "{self.component_names}."
                             .format(fail=fail, self=self,
                                     recurse_name=recurse_name))
        group = groupby(((child, rest)
                         for child, _, rest in (c.partition('.')
                                                for c in extra)),
                        lambda x: x[0])
        # we are into grand-children, can not be lazy!
        for child, cf_list in group:
            cpt = getattr(self, child)
            cpt.kind |= set_kind
            setattr(cpt,
                    recurse_name,
                    [c[1] for c in cf_list])

    @property
    def signal_names(self):
        warnings.warn("'signal_names' has been renamed 'component_names' for "
                      "clarity because it may include a mixture of Signals "
                      "and Devices -- any Components. This alias may be "
                      "removed in a future release of ophyd.", stacklevel=2)
        return self.component_names

    def summary(self):
        print(self._summary())

    def _summary(self):
        "Return a string summarizing the structure of the Device."
        desc = self.describe()
        config_desc = self.describe_configuration()
        read_attrs = self.read_attrs
        config_attrs = self.configuration_attrs
        used_attrs = set(read_attrs + config_attrs)
        extra_attrs = [a for a in self.component_names
                       if a not in used_attrs]
        hints = getattr(self, 'hints', {}).get('fields', [])

        def format_leaf(a):
            s = getattr(self, a)
            return '{:<20} {:<20}({!r})'.format(a, type(s).__name__,
                                                s.name)

        out = []
        out.append('data keys (* hints)')
        out.append('-------------------')
        for k in sorted(desc):
            out.append(('*' if k in hints else ' ') + k)
        out.append('')

        out.append('read attrs')
        out.append('----------')
        for a in read_attrs:
            out.append(format_leaf(a))

        out.append('')
        out.append('config keys')
        out.append('-----------')
        for k in sorted(config_desc):
            out.append(k)
        out.append('')

        out.append('configuration attrs')
        out.append('----------')
        for a in config_attrs:
            out.append(format_leaf(a))
        out.append('')

        out.append('Unused attrs')
        out.append('------------')
        for a in extra_attrs:
            out.append(format_leaf(a))
        return '\n'.join(out)

    def wait_for_connection(self, all_signals=False, timeout=2.0):
        '''Wait for signals to connect

        Parameters
        ----------
        all_signals : bool, optional
            Wait for all signals to connect (including lazy ones)
        timeout : float or None
            Overall timeout
        '''
        names = [attr for attr, cpt in self._sig_attrs.items()
                 if not cpt.lazy or all_signals]

        # Instantiate first to kickoff connection process
        signals = [getattr(self, name) for name in names]

        t0 = ttime.time()
        while timeout is None or (ttime.time() - t0) < timeout:
            connected = [sig.connected for sig in signals]
            if all(connected):
                return
            ttime.sleep(min((0.05, timeout / 10.0)))

        unconnected = ', '.join(self._get_unconnected())
        raise TimeoutError('Failed to connect to all signals: {}'
                           ''.format(unconnected))

    def _get_unconnected(self):
        '''Yields all of the signal pvnames or prefixes that are unconnected

        This recurses throughout the device hierarchy, only checking signals
        that have already been instantiated.
        '''
        for attr, sig in self.get_instantiated_signals():
            if sig.connected:
                continue

            if hasattr(sig, 'pvname'):
                prefix = sig.pvname
            else:
                prefix = sig.prefix

            yield '{} ({})'.format(attr, prefix)

    def get_instantiated_signals(self, *, attr_prefix=None):
        '''Yields all of the instantiated signals in a device hierarchy

        Parameters
        ----------
        attr_prefix : string, optional
            The attribute prefix. If None, defaults to self.name

        Yields
        ------
            (fully_qualified_attribute_name, signal_instance)
        '''
        if attr_prefix is None:
            attr_prefix = self.name

        for attr, sig in self._signals.items():
            # fully qualified attribute name from top-level device
            full_attr = '{}.{}'.format(attr_prefix, attr)
            if isinstance(sig, Device):
                yield from sig.get_instantiated_signals(attr_prefix=full_attr)
            else:
                yield full_attr, sig

    @property
    def connected(self):
        return all(signal.connected for name, signal in self._signals.items())

    def __getattr__(self, name):
        '''Get a component from a fully-qualified name

        As a reminder, __getattr__ is only called if a real attribute doesn't
        already exist, or a device component has yet to be instantiated.
        '''
        if '.' not in name:
            try:
                # Initial access of signal
                cpt = self._sig_attrs[name]
                return cpt.__get__(self, None)
            except KeyError:
                raise AttributeError(name)

        attr_names = name.split('.')
        try:
            attr = getattr(self, attr_names[0])
        except AttributeError:
            raise AttributeError('{} of {}'.format(attr_names[0], name))

        if len(attr_names) > 1:
            sub_attr_names = '.'.join(attr_names[1:])
            return getattr(attr, sub_attr_names)

        return attr

    @doc_annotation_forwarder(BlueskyInterface)
    def read(self):
        res = super().read()
        for component_name in self.component_names:
            # this might be lazy and get the Cpt
            component = _lazy_get(self, component_name)
            if component.kind & Kind.normal:
                # this forces us to get the real version
                component = getattr(self, component_name)
                res.update(component.read())
        return res

    def read_configuration(self) -> OrderedDictType[str, Dict[str, Any]]:
        """
        returns dictionary mapping names to (value, timestamp) pairs

        To control which fields are included, adjust the
        ``configuration_attrs`` list.
        """
        res = OrderedDict()
        for component_name in self.component_names:
            # this might be lazy and get the Cpt
            component = _lazy_get(self, component_name)
            if component.kind & Kind.config:
                # this forces us to get the real version
                component = getattr(self, component_name)
                res.update(component.read_configuration())
        return res

    @doc_annotation_forwarder(BlueskyInterface)
    def describe(self):
        res = super().describe()
        for component_name in self.component_names:
            component = _lazy_get(self, component_name)
            if component.kind & Kind.normal:
                component = getattr(self, component_name)
                res.update(component.describe())
        return res

    def describe_configuration(self) -> OrderedDictType[str, Dict[str, Any]]:
        """Provide schema & meta-data for :meth:`~BlueskyInterface.read_configuration`

        This keys in the `OrderedDict` this method returns must match the
        keys in the `OrderedDict` return by :meth:`~BlueskyInterface.read`.

        This provides schema related information, (ex shape, dtype), the
        source (ex PV name), and if available, units, limits, precision etc.

        Returns
        -------
        data_keys : OrderedDict
            The keys must be strings and the values must be dict-like
            with the ``event_model.event_descriptor.data_key`` schema.
        """
        res = OrderedDict()
        for component_name in self.component_names:
            component = _lazy_get(self, component_name)
            if component.kind & Kind.config:
                component = getattr(self, component_name)
                res.update(component.describe_configuration())
        return res

    @property
    def hints(self):
        fields = []
        for component_name in self.component_names:
            # Pick off the component's kind without instantiating it.
            kind = _lazy_get(self, component_name).kind
            if Kind.normal & kind:
                # OK, we have to instantiate it.
                component = getattr(self, component_name)
                c_hints = component.hints
                fields.extend(c_hints.get('fields', []))
        return {'fields': fields}

    @property
    def trigger_signals(self):
        names = [attr for attr, cpt in self._sig_attrs.items()
                 if cpt.trigger_value is not None]

        return [getattr(self, name) for name in names]

    def _done_acquiring(self, **kwargs):
        '''Call when acquisition has completed.'''
        self._run_subs(sub_type=self.SUB_ACQ_DONE,
                       success=True, **kwargs)

        self._reset_sub(self.SUB_ACQ_DONE)

    @doc_annotation_forwarder(BlueskyInterface)
    def trigger(self):
        """Start acquisition"""
        signals = self.trigger_signals
        if len(signals) > 1:
            raise NotImplementedError('More than one trigger signal is not '
                                      'currently supported')
        status = DeviceStatus(self)
        if not signals:
            status._finished()
            return status

        acq_signal, = signals

        self.subscribe(status._finished,
                       event_type=self.SUB_ACQ_DONE, run=False)

        def done_acquisition(**ignored_kwargs):
            # Keyword arguments are ignored here from the EpicsSignal
            # subscription, as the important part is that the put completion
            # has finished
            self._done_acquiring()

        acq_signal.put(1, wait=False, callback=done_acquisition)
        return status

    def stop(self, *, success=False):
        '''Stop the Device and all (instantiated) subdevices'''
        exc_list = []

        for attr in self._sub_devices:
            dev = getattr(self, attr)

            if not dev.connected:
                self.log.debug('stop: device %s (%s) is not connected; '
                               'skipping', attr, dev)
                continue

            try:
                dev.stop(success=success)
            except ExceptionBundle as ex:
                exc_list.extend([('{}.{}'.format(attr, sub_attr), ex)
                                 for sub_attr, ex in ex.exceptions.items()])
            except Exception as ex:
                exc_list.append((attr, ex))
                self.log.exception('Device %s (%s) stop failed', attr, dev)

        if exc_list:
            exc_info = '\n'.join('{} raised {!r}'.format(attr, ex)
                                 for attr, ex in exc_list)
            raise ExceptionBundle('{} exception(s) were raised during stop: \n'
                                  '{}'.format(len(exc_list), exc_info),
                                  exceptions=dict(exc_list))

    def get(self, **kwargs):
        '''Get the value of all components in the device

        Keyword arguments are passed onto each signal.get(). Components
        beginning with an underscore will not be included.
        '''
        values = {}
        for attr in self.component_names:
            if not attr.startswith('_'):
                signal = getattr(self, attr)
                values[attr] = signal.get(**kwargs)

        return self._device_tuple(**values)

    def put(self, dev_t, **kwargs):
        '''Put a value to all components of the device

        Keyword arguments are passed onto each signal.put()

        Parameters
        ----------
        dev_t : DeviceTuple or tuple
            The device tuple with the value(s) to put (see get_device_tuple)
        '''
        if not isinstance(dev_t, self._device_tuple):
            try:
                dev_t = self._device_tuple(dev_t)
            except TypeError as ex:
                raise ValueError('{}\n\tDevice tuple fields: {}'
                                 ''.format(ex, self._device_tuple._fields))

        for attr in self.component_names:
            value = getattr(dev_t, attr)
            signal = getattr(self, attr)
            signal.put(value, **kwargs)

    @classmethod
    def get_device_tuple(cls):
        '''The device tuple type associated with an Device class

        This is a tuple representing the full state of all components and
        dynamic device sub-components.
        '''
        return cls._device_tuple

    def configure(self,
                  d: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        '''Configure the device for something during a run

        This default implementation allows the user to change any of the
        `configuration_attrs`. Subclasses might override this to perform
        additional input validation, cleanup, etc.

        Parameters
        ----------
        d : dict
            The configuration dictionary. To specify the order that
            the changes should be made, use an OrderedDict.

        Returns
        -------
        (old, new) tuple of dictionaries
        Where old and new are pre- and post-configure configuration states.
        '''
        old = self.read_configuration()
        for key, val in d.items():
            if key not in self.configuration_attrs:
                # a little extra checking for a more specific error msg
                if key not in self.component_names:
                    raise ValueError("There is no signal named %s" % key)
                else:
                    raise ValueError("%s is not one of the "
                                     "configuration_fields, so it cannot be "
                                     "changed using configure" % key)
            set_and_wait(getattr(self, key), val)
        new = self.read_configuration()
        return old, new

    def _repr_info(self):
        yield ('prefix', self.prefix)
        yield from super()._repr_info()

        yield ('read_attrs', self.read_attrs)
        yield ('configuration_attrs', self.configuration_attrs)

    class OphydAttrList(MutableSequence):
        """list proxy to migrate away from Device.read_attrs and Device.config_attrs


        """
        def __init__(self, device, kind, remove_kind, recurse_key):
            self._kind = kind
            self._remove_kind = remove_kind
            self._parent = device
            self._recurse_key = recurse_key

        def __internal_list(self):

            children = [c for c in self._parent.component_names
                        if _lazy_get(self._parent, c).kind & self._kind]
            out = []
            for c in children:
                cmpt = getattr(self._parent, c)
                out.append(c)
                if hasattr(cmpt, self._recurse_key):
                    out.extend('.'.join([c, v]) for v in
                               getattr(cmpt, self._recurse_key, []))

            return out

        def __getitem__(self, key):
            return self.__internal_list()[key]

        def __setitem__(self, key, val):
            raise NotImplemented

        def __delitem__(self, key):
            o = self.__internal_list()[key]
            if not isinstance(key, slice):
                o = [o]
            for k in o:
                getattr(self._parent, k).kind &= ~self._remove_kind

        def __len__(self):
            return len(self.__internal_list())

        def insert(self, index, object):
            getattr(self._parent, object).kind |= self._kind

        def remove(self, value):
            getattr(self._parent, value).kind &= ~self._remove_kind

        def __contains__(self, value):
            return value in self.__internal_list()

        def __iter__(self):
            yield from self.__internal_list()

        def __eq__(self, other):
            return list(self) == other

        def __repr__(self):
            return repr(list(self))

        def __add__(self, other):
            return list(self) + list(other)


@contextlib.contextmanager
def kind_context(kind):
    yield functools.partial(Component, kind=kind)


def _lazy_get(parent, name):
    return parent._signals.get(name, parent._initial_state[name])


def _ensure_kind(k):
    return getattr(Kind, k.lower()) if isinstance(k, str) else k



'''PV names for tests'''
import os

epics_base = os.environ.get('EPICS_BASE', '/usr/lib/epics')

motor_recs = ['XF:31IDA-OP{Tbl-Ax:X1}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X2}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X3}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X4}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X5}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X6}Mtr',
              ]

fake_motors = [{'readback': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}-I',
                'setpoint': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}-SP',
                'moving': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}Sts:Moving-Sts',
                'actuate': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}Cmd:Go-Cmd.PROC',
                'stop': 'XF:31IDA-OP{Tbl-Ax:FakeMtr}Cmd:Stop-Cmd.PROC',
                },

               ]

fake_sensors = ['XF:31IDA-BI{Dev:1}E-I',
                'XF:31IDA-BI{Dev:2}E-I',
                'XF:31IDA-BI{Dev:3}E-I',
                'XF:31IDA-BI{Dev:4}E-I',
                'XF:31IDA-BI{Dev:5}E-I',
                'XF:31IDA-BI{Dev:6}E-I',
                ]

ad_plugins = {'image': ['image1:', ],
              'roi': ['ROI1:', 'ROI2:', 'ROI3:', 'ROI4:'],
              'file': ['netCDF1:', 'TIFF1:', 'JPEG1:', 'Nexus1:', 'HDF1:', 'Magick1:', ],
              'proc': ['Proc1:', ],
              'stats': ['Stats1:', 'Stats2:', 'Stats3:', 'Stats4:', 'Stats5:', ],
              'overlay': [('Over1:', 1, 8), ],
              'trans': ['Trans1:', ],
              'cc': ['CC1:', 'CC2:', ],
              }

sim_areadetector = [{'prefix': 'XF:31IDA-BI{Cam:Tbl}',
                     'cam': 'cam1:',
                     },
                    {'prefix': 'XF:31IDA-BI{Cam:Tbl}',
                     'cam': 'cam2:',
                     }
                    ]

# this is a real instrument - it is not currently used, but be cautious!
# scalers = ['XF:23ID2-ES{Sclr:1}']




import logging

from collections import OrderedDict

from .status import DeviceStatus
from .signal import (Signal, EpicsSignal, EpicsSignalRO)
from .device import (Device, Component as C, DynamicDeviceComponent as DDC,
                     Staged, BlueskyInterface, ALL_COMPONENTS, Kind)
from .areadetector import EpicsSignalWithRBV as SignalWithRBV


logger = logging.getLogger(__name__)


class ROI(Device):

    # 'name' is not an allowed attribute
    label = C(EpicsSignal, 'NM', lazy=True)
    count = C(EpicsSignalRO, '', lazy=True)
    net_count = C(EpicsSignalRO, 'N', lazy=True)
    preset_count = C(EpicsSignal, 'P', lazy=True)
    is_preset = C(EpicsSignal, 'IP', lazy=True)
    bkgnd_chans = C(EpicsSignal, 'BG', lazy=True)
    hi_chan = C(EpicsSignal, 'HI', lazy=True)
    lo_chan = C(EpicsSignal, 'LO', lazy=True)

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 name=None, parent=None, **kwargs):

        super().__init__(prefix, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs,
                         name=name, parent=parent, **kwargs)


def add_rois(range_, **kwargs):
    '''Add one or more ROIs to an MCA instance

       Parameters
       ----------
       range_ : sequence of ints
           Must be be in the set [0,31]

       By default, an EpicsMCA is initialized with all 32 rois.
       These provide the following Components as EpicsSignals (N=[0,31]):
       EpicsMCA.rois.roiN.(label,count,net_count,preset_cnt, is_preset,
       bkgnd_chans, hi_chan, lo_chan)
       '''
    defn = OrderedDict()

    for roi in range_:
        if not (0 <= roi < 32):
            raise ValueError('roi must be in the set [0,31]')

        attr = 'roi{}'.format(roi)
        defn[attr] = (ROI, '.R{}'.format(roi), kwargs)

    return defn


class EpicsMCARecord(Device):
    '''SynApps MCA Record interface'''
    stop_signal = C(EpicsSignal, '.STOP', kind='omitted')
    preset_real_time = C(EpicsSignal, '.PRTM', kind=Kind.config | Kind.normal)
    preset_live_time = C(EpicsSignal, '.PLTM', kind='omitted')
    elapsed_real_time = C(EpicsSignalRO, '.ERTM')
    elapsed_live_time = C(EpicsSignalRO, '.ELTM', kind='omitted')

    spectrum = C(EpicsSignalRO, '.VAL')
    background = C(EpicsSignalRO, '.BG', kind='omitted')
    mode = C(EpicsSignal, '.MODE', string=True, kind='omitted')

    rois = DDC(add_rois(range(0, 32), kind='omitted'), kind='omitted')

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # could arguably be made a configuration_attr instead...
        self.stage_sigs['mode'] = 'PHA'

    def stop(self, *, success=False):
        self.stop_signal.put(1)


class EpicsMCA(EpicsMCARecord):
    '''mca records with extras from mca.db'''
    start = C(EpicsSignal, 'Start', kind='omitted')
    stop_signal = C(EpicsSignal, 'Stop', kind='omitted')
    erase = C(EpicsSignal, 'Erase', kind='omitted')
    erase_start = C(EpicsSignal, 'EraseStart', trigger_value=1, kind='omitted')

    check_acquiring = C(EpicsSignal, 'CheckACQG', kind='omitted')
    client_wait = C(EpicsSignal, 'ClientWait', kind='omitted')
    enable_wait = C(EpicsSignal, 'EnableWait', kind='omitted')
    force_read = C(EpicsSignal, 'Read', kind='omitted')
    set_client_wait = C(EpicsSignal, 'SetClientWait', kind='omitted')
    status = C(EpicsSignal, 'Status', kind='omitted')
    when_acq_stops = C(EpicsSignal, 'WhenAcqStops', kind='omitted')
    why1 = C(EpicsSignal, 'Why1', kind='omitted')
    why2 = C(EpicsSignal, 'Why2', kind='omitted')
    why3 = C(EpicsSignal, 'Why3', kind='omitted')
    why4 = C(EpicsSignal, 'Why4', kind='omitted')


class EpicsMCAReadNotify(EpicsMCARecord):
    '''mca record with extras from mcaReadNotify.db'''
    start = C(EpicsSignal, 'Start', kind='omitted')
    stop_signal = C(EpicsSignal, 'Stop', kind='omitted')
    erase = C(EpicsSignal, 'Erase', kind='omitted')
    erase_start = C(EpicsSignal, 'EraseStart', trigger_value=1, kind='omitted')

    check_acquiring = C(EpicsSignal, 'CheckACQG', kind='omitted')
    client_wait = C(EpicsSignal, 'ClientWait', kind='omitted')
    enable_wait = C(EpicsSignal, 'EnableWait', kind='omitted')
    force_read = C(EpicsSignal, 'Read', kind='omitted')
    set_client_wait = C(EpicsSignal, 'SetClientWait', kind='omitted')
    status = C(EpicsSignal, 'Status', kind='omitted')


class EpicsMCACallback(Device):
    '''Callback-related signals for MCA devices'''
    read_callback = C(EpicsSignal, 'ReadCallback')
    read_data_once = C(EpicsSignal, 'ReadDataOnce')
    read_status_once = C(EpicsSignal, 'ReadStatusOnce')
    collect_data = C(EpicsSignal, 'CollectData')


class EpicsDXP(Device):
    '''All high-level DXP parameters for each channel'''
    preset_mode = C(EpicsSignal, 'PresetMode', string=True)

    live_time_output = C(SignalWithRBV, 'LiveTimeOutput', string=True)
    elapsed_live_time = C(EpicsSignal, 'ElapsedLiveTime')
    elapsed_real_time = C(EpicsSignal, 'ElapsedRealTime')
    elapsed_trigger_live_time = C(EpicsSignal, 'ElapsedTriggerLiveTime')

    # Trigger Filter PVs
    trigger_peaking_time = C(SignalWithRBV, 'TriggerPeakingTime')
    trigger_threshold = C(SignalWithRBV, 'TriggerThreshold')
    trigger_gap_time = C(SignalWithRBV, 'TriggerGapTime')
    trigger_output = C(SignalWithRBV, 'TriggerOutput', string=True)
    max_width = C(SignalWithRBV, 'MaxWidth')

    # Energy Filter PVs
    peaking_time = C(SignalWithRBV, 'PeakingTime')
    energy_threshold = C(SignalWithRBV, 'EnergyThreshold')
    gap_time = C(SignalWithRBV, 'GapTime')

    # Baseline PVs
    baseline_cut_percent = C(SignalWithRBV, 'BaselineCutPercent')
    baseline_cut_enable = C(SignalWithRBV, 'BaselineCutEnable')
    baseline_filter_length = C(SignalWithRBV, 'BaselineFilterLength')
    baseline_threshold = C(SignalWithRBV, 'BaselineThreshold')
    baseline_energy_array = C(EpicsSignal, 'BaselineEnergyArray')
    baseline_histogram = C(EpicsSignal, 'BaselineHistogram')
    baseline_threshold = C(SignalWithRBV, 'BaselineThreshold')

    # Misc PVs
    preamp_gain = C(SignalWithRBV, 'PreampGain')
    detector_polarity = C(SignalWithRBV, 'DetectorPolarity')
    reset_delay = C(SignalWithRBV, 'ResetDelay')
    decay_time = C(SignalWithRBV, 'DecayTime')
    max_energy = C(SignalWithRBV, 'MaxEnergy')
    adc_percent_rule = C(SignalWithRBV, 'ADCPercentRule')
    max_width = C(SignalWithRBV, 'MaxWidth')

    # read-only diagnostics
    triggers = C(EpicsSignalRO, 'Triggers', lazy=True)
    events = C(EpicsSignalRO, 'Events', lazy=True)
    overflows = C(EpicsSignalRO, 'Overflows', lazy=True)
    underflows = C(EpicsSignalRO, 'Underflows', lazy=True)
    input_count_rate = C(EpicsSignalRO, 'InputCountRate', lazy=True)
    output_count_rate = C(EpicsSignalRO, 'OutputCountRate', lazy=True)

    mca_bin_width = C(EpicsSignalRO, 'MCABinWidth_RBV')
    calibration_energy = C(EpicsSignalRO, 'CalibrationEnergy_RBV')
    current_pixel = C(EpicsSignal, 'CurrentPixel')
    dynamic_range = C(EpicsSignalRO, 'DynamicRange_RBV')

    # Preset options
    preset_events = C(SignalWithRBV, 'PresetEvents')
    preset_mode = C(SignalWithRBV, 'PresetMode', string=True)
    preset_triggers = C(SignalWithRBV, 'PresetTriggers')

    # Trace options
    trace_data = C(EpicsSignal, 'TraceData')
    trace_mode = C(SignalWithRBV, 'TraceMode', string=True)
    trace_time_array = C(EpicsSignal, 'TraceTimeArray')
    trace_time = C(SignalWithRBV, 'TraceTime')


class EpicsDXPLowLevelParameter(Device):
    param_name = C(EpicsSignal, 'Name')
    value = C(SignalWithRBV, 'Val')


class EpicsDXPLowLevel(Device):
    num_low_level_params = C(EpicsSignal, 'NumLLParams')
    read_low_level_params = C(EpicsSignal, 'ReadLLParams')

    parameter_prefix = 'LL{}'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parameter_cache = {}

    def get_low_level_parameter(self, index):
        '''Get a DXP low level parameter

        Parameters
        ----------
        index : int
            In the range of [0, 229]

        Returns
        -------
        param : EpicsDXPLowLevelParameter
        '''
        try:
            return self._parameter_cache[index]
        except KeyError:
            pass

        prefix = '{}{}'.format(self.prefix, self.parameter_prefix)
        name = '{}_param{}'.format(self.name, index)
        param = EpicsDXPLowLevelParameter(prefix, name=name)
        self._parameter_cache[index] = param
        return param


class EpicsDXPMapping(Device):
    apply = C(EpicsSignal, 'Apply')
    auto_apply = C(SignalWithRBV, 'AutoApply')
    auto_pixels_per_buffer = C(SignalWithRBV, 'AutoPixelsPerBuffer')
    buffer_size = C(EpicsSignalRO, 'BufferSize_RBV')
    collect_mode = C(SignalWithRBV, 'CollectMode')
    ignore_gate = C(SignalWithRBV, 'IgnoreGate')
    input_logic_polarity = C(SignalWithRBV, 'InputLogicPolarity')
    list_mode = C(SignalWithRBV, 'ListMode')
    mbytes_read = C(EpicsSignalRO, 'MBytesRead_RBV')
    next_pixel = C(EpicsSignal, 'NextPixel')
    pixel_advance_mode = C(SignalWithRBV, 'PixelAdvanceMode')
    pixels_per_buffer = C(SignalWithRBV, 'PixelsPerBuffer')
    pixels_per_run = C(SignalWithRBV, 'PixelsPerRun')
    read_rate = C(EpicsSignalRO, 'ReadRate_RBV')
    sync_count = C(SignalWithRBV, 'SyncCount')


class EpicsDXPBaseSystem(Device):
    channel_advance = C(EpicsSignal, 'ChannelAdvance')
    client_wait = C(EpicsSignal, 'ClientWait')
    dwell = C(EpicsSignal, 'Dwell')
    max_scas = C(EpicsSignal, 'MaxSCAs')
    num_scas = C(SignalWithRBV, 'NumSCAs')
    poll_time = C(SignalWithRBV, 'PollTime')
    prescale = C(EpicsSignal, 'Prescale')
    save_system = C(SignalWithRBV, 'SaveSystem')
    save_system_file = C(EpicsSignal, 'SaveSystemFile')
    set_client_wait = C(EpicsSignal, 'SetClientWait')


class EpicsDXPMultiElementSystem(EpicsDXPBaseSystem):
    # Preset info
    preset_events = C(EpicsSignal, 'PresetEvents')
    preset_live_time = C(EpicsSignal, 'PresetLive')
    preset_real_time = C(EpicsSignal, 'PresetReal')
    preset_mode = C(EpicsSignal, 'PresetMode', string=True)
    preset_triggers = C(EpicsSignal, 'PresetTriggers')

    # Acquisition
    erase_all = C(EpicsSignal, 'EraseAll')
    erase_start = C(EpicsSignal, 'EraseStart', trigger_value=1)
    start_all = C(EpicsSignal, 'StartAll')
    stop_all = C(EpicsSignal, 'StopAll')

    # Status
    set_acquire_busy = C(EpicsSignal, 'SetAcquireBusy')
    acquire_busy = C(EpicsSignal, 'AcquireBusy')
    status_all = C(EpicsSignal, 'StatusAll')
    status_all_once = C(EpicsSignal, 'StatusAllOnce')
    acquiring = C(EpicsSignal, 'Acquiring')

    # Reading
    read_baseline_histograms = C(EpicsSignal, 'ReadBaselineHistograms')
    read_all = C(EpicsSignal, 'ReadAll')
    read_all_once = C(EpicsSignal, 'ReadAllOnce')

    # As a debugging note, if snl_connected is not '1', your IOC is
    # misconfigured:
    snl_connected = C(EpicsSignal, 'SNL_Connected')

    # Copying to individual elements
    copy_adcp_ercent_rule = C(EpicsSignal, 'CopyADCPercentRule')
    copy_baseline_cut_enable = C(EpicsSignal, 'CopyBaselineCutEnable')
    copy_baseline_cut_percent = C(EpicsSignal, 'CopyBaselineCutPercent')
    copy_baseline_filter_length = C(EpicsSignal, 'CopyBaselineFilterLength')
    copy_baseline_threshold = C(EpicsSignal, 'CopyBaselineThreshold')
    copy_decay_time = C(EpicsSignal, 'CopyDecayTime')
    copy_detector_polarity = C(EpicsSignal, 'CopyDetectorPolarity')
    copy_energy_threshold = C(EpicsSignal, 'CopyEnergyThreshold')
    copy_gap_time = C(EpicsSignal, 'CopyGapTime')
    copy_max_energy = C(EpicsSignal, 'CopyMaxEnergy')
    copy_max_width = C(EpicsSignal, 'CopyMaxWidth')
    copy_peaking_time = C(EpicsSignal, 'CopyPeakingTime')
    copy_preamp_gain = C(EpicsSignal, 'CopyPreampGain')
    copy_roic_hannel = C(EpicsSignal, 'CopyROIChannel')
    copy_roie_nergy = C(EpicsSignal, 'CopyROIEnergy')
    copy_roi_sca = C(EpicsSignal, 'CopyROI_SCA')
    copy_reset_delay = C(EpicsSignal, 'CopyResetDelay')
    copy_trigger_gap_time = C(EpicsSignal, 'CopyTriggerGapTime')
    copy_trigger_peaking_time = C(EpicsSignal, 'CopyTriggerPeakingTime')
    copy_trigger_threshold = C(EpicsSignal, 'CopyTriggerThreshold')

    # do_* executes the process:
    do_read_all = C(EpicsSignal, 'DoReadAll')
    do_read_baseline_histograms = C(EpicsSignal, 'DoReadBaselineHistograms')
    do_read_traces = C(EpicsSignal, 'DoReadTraces')
    do_status_all = C(EpicsSignal, 'DoStatusAll')

    # Time
    dead_time = C(EpicsSignal, 'DeadTime')
    elapsed_live = C(EpicsSignal, 'ElapsedLive')
    elapsed_real = C(EpicsSignal, 'ElapsedReal')
    idead_time = C(EpicsSignal, 'IDeadTime')

    # low-level
    read_low_level_params = C(EpicsSignal, 'ReadLLParams')

    # Traces
    read_traces = C(EpicsSignal, 'ReadTraces')
    trace_modes = C(EpicsSignal, 'TraceModes', string=True)
    trace_times = C(EpicsSignal, 'TraceTimes')


class SaturnMCA(EpicsMCA, EpicsMCACallback):
    pass


class SaturnDXP(EpicsDXP, EpicsDXPLowLevel):
    pass


class Saturn(EpicsDXPBaseSystem):
    '''DXP Saturn with 1 channel example'''
    dxp = C(SaturnDXP, 'dxp1:')
    mca = C(SaturnMCA, 'mca1')


class MercuryDXP(EpicsDXP, EpicsDXPLowLevel):
    pass


class Mercury1(EpicsDXPMultiElementSystem):
    '''DXP Mercury with 1 channel example'''
    dxp = C(MercuryDXP, 'dxp1:')
    mca = C(EpicsMCARecord, 'mca1')


class SoftDXPTrigger(BlueskyInterface):
    '''Simple soft trigger for DXP devices

    Parameters
    ----------
    count_signal : str, optional
        Signal to set acquisition time (default: 'preset_real_time')
    preset_mode : str, optional
        Default preset mode for the stage signals (default: 'Real time')
    mode_signal : str, optional
        Preset mode signal attribute (default 'preset_mode')
    stop_signal : str, optional
        Stop signal attribute (default 'stop_all')
    '''

    count_time = C(Signal, value=None, doc='bluesky count time')

    def __init__(self, *args, count_signal='preset_real_time',
                 stop_signal='stop_all', mode_signal='preset_mode',
                 preset_mode='Real time',
                 **kwargs):
        super().__init__(*args, **kwargs)
        self._status = None
        self._count_signal = getattr(self, count_signal)

        stop_signal = getattr(self, stop_signal)
        self.stage_sigs[stop_signal] = 1

        mode_signal = getattr(self, mode_signal)
        self.stage_sigs[mode_signal] = preset_mode

    def stage(self):
        if self.count_time.get() is None:
            # remove count_time from the stage signals if count_time unset
            try:
                del self.stage_sigs[self._count_signal]
            except KeyError:
                pass
        else:
            self.stage_sigs[self._count_signal] = self.count_time.get()

        super().stage()



import os
import warnings


warnings.warn("This module is deprecated in ophyd 1.1 and "
              "will be removed in 1.2. "
              "use the cl attribute at the top level instead",
              stacklevel=2)

__all__ = ['setup',
           'caput', 'caget',
           'get_pv',
           'pv_form',
           'thread_class']
_cl = os.environ.get('OPHYD_CONTROL_LAYER', '').lower()

if _cl == 'caproto':
    from ._caproto_shim import setup, caput, caget, get_pv, pv_form, thread_class
elif '_cl' == 'dummy':
    from ._dummy_shim import setup, caput, caget, get_pv, pv_form, thread_class
else:
    try:
        from ._pyepics_shim import setup, caput, caget, get_pv, pv_form, thread_class
    except ImportError:
        from ._dummy_shim import setup, caput, caget, get_pv, pv_form, thread_class
del _cl



from itertools import count

import time
import logging

from .status import (StatusBase, MoveStatus, DeviceStatus)

try:
    from enum import IntFlag
except ImportError:
    # we must be in python 3.5
    from .utils._backport_enum import IntFlag


class Kind(IntFlag):
    """
    This is used in the .kind attribute of all OphydObj (Signals, Devices).

    A Device examines its components' .kind atttribute to decide whether to
    traverse it in read(), read_configuration(), or neither. Additionally, if
    decides whether to include its name in `.hints['fields']`.
    """
    omitted = 0
    normal = 1
    config = 2
    hinted = 5  # Notice that bool(hinted & normal) is True.


class UnknownSubscription(KeyError):
    "Subclass of KeyError.  Raised for unknown event type"
    ...


class OphydObject:
    '''The base class for all objects in Ophyd

    Handles:

      * Subscription/callback mechanism

    Parameters
    ----------
    name : str, optional
        The name of the object.
    parent : parent, optional
        The object's parent, if it exists in a hierarchy
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.

    Attributes
    ----------
    name
    '''

    _default_sub = None

    def __init__(self, *, name=None, parent=None, labels=None,
                 kind=None):
        if labels is None:
            labels = set()
        self._ophyd_labels_ = set(labels)
        if kind is None:
            kind = Kind.normal
        self.kind = kind

        super().__init__()

        # base name and ref to parent, these go with properties
        if name is None:
            name = ''
        if not isinstance(name, str):
            raise ValueError("name must be a string.")
        self._name = name
        self._parent = parent

        self.subscriptions = {getattr(self, k)
                              for k in dir(type(self))
                              if (k.startswith('SUB') or
                                  k.startswith('_SUB'))}

        # dictionary of wrapped callbacks
        self._callbacks = {k: {} for k in self.subscriptions}
        # this is to maintain api on clear_sub
        self._unwrapped_callbacks = {k: {} for k in self.subscriptions}
        # map cid -> back to which event it is in
        self._cid_to_event_mapping = dict()
        # cache of last inputs to _run_subs, the semi-private way
        # to trigger the callbacks for a given subscription to be run
        self._args_cache = {k: None for k in self.subscriptions}
        # count of subscriptions we have handed out, used to give unique ids
        self._cb_count = count()
        # Create logger name from parent or from module class
        if self.parent:
            base_log = self.parent.log.name
            name = self.name.lstrip(self.parent.name + '_')
        else:
            base_log = self.__class__.__module__
            name = self.name
        # Instantiate logger
        self.log = logging.getLogger(base_log + '.' + name)

    def _validate_kind(self, val):
        if isinstance(val, str):
            val = getattr(Kind, val.lower())
        return val

    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(self, val):
        self._kind = self._validate_kind(val)

    @property
    def name(self):
        '''name of the device'''
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def connected(self):
        '''If the device is connected.

        Subclasses should override this'''
        return True

    @property
    def parent(self):
        '''The parent of the ophyd object.

        If at the top of its hierarchy, `parent` will be None
        '''
        return self._parent

    @property
    def root(self):
        "Walk parents to find ultimate ancestor (parent's parent...)."
        root = self
        while True:
            if root.parent is None:
                return root
            root = root.parent

    @property
    def report(self):
        '''A report on the object.'''
        return {}

    @property
    def event_types(self):
        '''Events that can be subscribed to via `obj.subscribe`
        '''
        return tuple(self.subscriptions)

    def _run_subs(self, *args, sub_type, **kwargs):
        '''Run a set of subscription callbacks

        Only the kwarg ``sub_type`` is required, indicating
        the type of callback to perform. All other positional arguments
        and kwargs are passed directly to the callback function.

        The host object will be injected into kwargs as 'obj' unless that key
        already exists.

        If the `timestamp` is None, then it will be replaced by the current
        time.

        No exceptions are raised if the callback functions fail.
        '''
        if sub_type not in self.subscriptions:
            raise UnknownSubscription(
                "Unknown subscription {}, must be one of {!r}"
                .format(sub_type, self.subscriptions))

        kwargs['sub_type'] = sub_type
        # Guarantee that the object will be in the kwargs
        kwargs.setdefault('obj', self)

        # And if a timestamp key exists, but isn't filled -- supply it with
        # a new timestamp
        if 'timestamp' in kwargs and kwargs['timestamp'] is None:
            kwargs['timestamp'] = time.time()

        # Shallow-copy the callback arguments for replaying the
        # callback at a later time (e.g., when a new subscription is made)
        self._args_cache[sub_type] = (tuple(args), dict(kwargs))

        for cb in list(self._callbacks[sub_type].values()):
            cb(*args, **kwargs)

    def subscribe(self, cb, event_type=None, run=True):
        '''Subscribe to events this event_type generates.

        The callback will be called as ``cb(*args, **kwargs)`` with
        the values passed to `_run_subs` with the following additional keys:

           sub_type : the string value of the event_type
           obj : the host object, added if 'obj' not already in kwargs

        if the key 'timestamp' is in kwargs _and_ is None, then it will
        be replaced with the current time before running the callback.

        The ``*args``, ``**kwargs`` passed to _run_subs will be cached as
        shallow copies, be aware of passing in mutable data.

        .. warning::

           If the callback raises any exceptions when run they will be
           silently ignored.

        Parameters
        ----------
        cb : callable
            A callable function (that takes kwargs) to be run when the event is
            generated.  The expected signature is ::

              def cb(*args, obj: OphydObject, sub_type: str, **kwargs) -> None:

            The exact args/kwargs passed are whatever are passed to
            ``_run_subs``
        event_type : str, optional
            The name of the event to subscribe to (if None, defaults to
            the default sub for the instance - obj._default_sub)

            This maps to the ``sub_type`` kwargs in `_run_subs`
        run : bool, optional
            Run the callback now

        See Also
        --------
        clear_sub, _run_subs

        Returns
        -------
        cid : int
            id of callback, can be passed to `unsubscribe` to remove the
            callback

        '''
        if not callable(cb):
            raise ValueError("cb must be callable")
        # do default event type
        if event_type is None:
            # warnings.warn("Please specify which call back you wish to "
            #               "attach to defaulting to {}"
            #               .format(self._default_sub), stacklevel=2)
            event_type = self._default_sub

        if event_type is None:
            raise ValueError('Subscription type not set and object {} of class'
                             ' {} has no default subscription set'
                             ''.format(self.name, self.__class__.__name__))

        # check that this is a valid event type
        if event_type not in self.subscriptions:
            raise UnknownSubscription(
                "Unknown subscription {}, must be one of {!r}"
                .format(event_type, self.subscriptions))

        # wrapper for callback to snarf exceptions
        def wrap_cb(cb):
            def inner(*args, **kwargs):
                try:
                    cb(*args, **kwargs)
                except Exception:
                    sub_type = kwargs['sub_type']
                    self.log.exception('Subscription %s callback '\
                                       'exception (%s)',
                                       sub_type, self)
            return inner
        # get next cid
        cid = next(self._cb_count)
        wrapped = wrap_cb(cb)
        self._unwrapped_callbacks[event_type][cid] = cb
        self._callbacks[event_type][cid] = wrapped
        self._cid_to_event_mapping[cid] = event_type

        if run:
            cached = self._args_cache[event_type]
            if cached is not None:
                args, kwargs = cached
                wrapped(*args, **kwargs)

        return cid

    def _reset_sub(self, event_type):
        '''Remove all subscriptions in an event type'''
        self._callbacks[event_type].clear()
        self._unwrapped_callbacks[event_type].clear()

    def clear_sub(self, cb, event_type=None):
        '''Remove a subscription, given the original callback function

        See also :meth:`subscribe`, :meth:`unsubscribe`

        Parameters
        ----------
        cb : callable
            The callback
        event_type : str, optional
            The event to unsubscribe from (if None, removes it from all event
            types)
        '''
        if event_type is None:
            event_types = self.event_types
        else:
            event_types = [event_type]
        cid_list = []
        for et in event_types:
            for cid, target in self._unwrapped_callbacks[et].items():
                if cb == target:
                    cid_list.append(cid)
        for cid in cid_list:
            self.unsubscribe(cid)

    def unsubscribe(self, cid):
        """Remove a subscription

        See also :meth:`subscribe`, :meth:`clear_sub`

        Parameters
        ----------
        cid : int
           token return by :meth:`subscribe`
        """
        ev_type = self._cid_to_event_mapping.pop(cid, None)
        if ev_type is None:
            return
        del self._unwrapped_callbacks[ev_type][cid]
        del self._callbacks[ev_type][cid]

    def unsubscribe_all(self):
        for ev_type in self._callbacks:
            self._reset_sub(ev_type)

    def check_value(self, value, **kwargs):
        '''Check if the value is valid for this object

        This function does no normalization, but may raise if the
        value is invalid.

        Raises
        ------
        ValueError
        '''
        pass

    def __repr__(self):
        info = self._repr_info()
        info = ', '.join('{}={!r}'.format(key, value) for key, value in info)
        return '{}({})'.format(self.__class__.__name__, info)

    def _repr_info(self):
        if self.name is not None:
            yield ('name', self.name)

        if self._parent is not None:
            yield ('parent', self.parent.name)

    def __copy__(self):
        info = dict(self._repr_info())
        return self.__class__(**info)



import logging
import time
from collections import OrderedDict
import functools
from .ophydobj import OphydObject, Kind
from .status import (MoveStatus, wait as status_wait, StatusBase)
from .utils.epics_pvs import (data_type, data_shape)
from typing import Any, Callable
logger = logging.getLogger(__name__)


class PositionerBase(OphydObject):
    '''The positioner base class

    Subclass from this to implement your own positioners.

    .. note ::

       Subclasses should add an additional 'wait' keyword argument on
       the move method. The `MoveStatus` object returned from
       `PositionerBase` can then be waited on after the subclass
       finishes the motion configuration.

    '''

    SUB_START = 'start_moving'
    SUB_DONE = 'done_moving'
    SUB_READBACK = 'readback'
    _SUB_REQ_DONE = '_req_done'  # requested move finished subscription
    _default_sub = SUB_READBACK

    def __init__(self, *, name=None, parent=None, settle_time=0.0,
                 timeout=None, **kwargs):
        super().__init__(name=name, parent=parent, **kwargs)

        self._started_moving = False
        self._moving = False
        self._position = None
        self._settle_time = settle_time
        self._timeout = timeout

    # High level
    def set(self, new_position: Any, *,
            timeout: float=None,
            moved_cb: Callable =None,
            wait: bool =False) -> StatusBase:
        """Set a value and return a Status object


        Parameters
        ----------
        new_position : object

            The input here is whatever the device requires (this
            should be over-ridden by the implementation.  For example
            a motor would take a float, a shutter the strings {'Open',
            'Close'}, and a goineometer (h, k, l) tuples

        timeout : float, optional

            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        moved_cb : callable, optional
            Deprecated

            Call this callback when movement has finished. This callback
            must accept one keyword argument: 'obj' which will be set to
            this positioner instance.

        wait : bool, optional
            Deprecated

            If the method should block until the Status object reports
            it is done.

            Defaults to False

        Returns
        -------
        status : StatusBase
            Status object to indicate when the motion / set is done.
        """
        return self.move(new_position, wait=wait, moved_cb=moved_cb,
                         timeout=timeout)

    def stop(self, *, success: bool =False):
        '''Stops motion.

        Sub-classes must extend this method to _actually_ stop the device.

        Parameters
        ----------
        success : bool, optional
            If the move should be considered a success despite the stop.

            Defaults to False
        '''
        self._done_moving(success=success)

    # Suggested properties
    @property
    def settle_time(self):
        '''Amount of time to wait after moves to report status completion'''
        return self._settle_time

    @settle_time.setter
    def settle_time(self, settle_time):
        self._settle_time = settle_time

    @property
    def timeout(self):
        '''Amount of time to wait before to considering a motion as failed'''
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        if timeout is None:
            self._timeout = None
        else:
            self._timeout = float(timeout)

    # low level
    @property
    def report(self):
        rep = super().report
        rep['position'] = self.position
        return rep

    @property
    def egu(self):
        '''The engineering units (EGU) for positions'''
        raise NotImplementedError('Subclass must implement egu')

    @property
    def limits(self):
        return (0, 0)

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def move(self, position, moved_cb=None, timeout=None):
        '''Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback
            must accept one keyword argument: 'obj' which will be set to
            this positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        '''

        if timeout is None:
            timeout = self._timeout

        self.check_value(position)

        self._run_subs(sub_type=self._SUB_REQ_DONE, success=False)
        self._reset_sub(self._SUB_REQ_DONE)

        status = MoveStatus(self, position, timeout=timeout,
                            settle_time=self._settle_time)

        if moved_cb is not None:
            status.add_callback(functools.partial(moved_cb, obj=self))
            # the status object will run this callback when finished

        self.subscribe(status._finished, event_type=self._SUB_REQ_DONE,
                       run=False)

        return status

    def _done_moving(self, success=True, timestamp=None, value=None, **kwargs):
        '''Call when motion has completed.  Runs ``SUB_DONE`` subscription.'''
        if success:
            self._run_subs(sub_type=self.SUB_DONE, timestamp=timestamp,
                           value=value)

        self._run_subs(sub_type=self._SUB_REQ_DONE, success=success,
                       timestamp=timestamp)
        self._reset_sub(self._SUB_REQ_DONE)

    @property
    def position(self):
        '''The current position of the motor in its engineering units

        Returns
        -------
        position : any
        '''
        return self._position

    def _set_position(self, value, **kwargs):
        '''Set the current internal position, run the readback subscription'''
        self._position = value

        timestamp = kwargs.pop('timestamp', time.time())
        self._run_subs(sub_type=self.SUB_READBACK, timestamp=timestamp,
                       value=value, **kwargs)

    @property
    def moving(self):
        '''Whether or not the motor is moving

        Returns
        -------
        moving : bool
        '''
        return self._moving

    def _repr_info(self):
        yield from super()._repr_info()
        yield ('settle_time', self._settle_time)
        yield ('timeout', self._timeout)

    @property
    def hints(self):
        if (~Kind.normal & Kind.hinted) & self.kind:
            return {'fields': [self.name]}
        else:
            return {'fields': []}


class SoftPositioner(PositionerBase):
    '''A positioner which does not communicate with any hardware

    SoftPositioner 'moves' immediately to the target position when commanded to
    do so.

    Parameters
    ----------
    limits : (low_limit, high_limit)
        Soft limits to use
    egu : str, optional
        Engineering units (EGU) for a position
    source : str, optional
        Metadata indicating the source of this positioner's position. Defaults
        to 'computed'
    '''

    def __init__(self, *, egu='', limits=None, source='computed',
                 init_pos=None,
                 **kwargs):
        super().__init__(**kwargs)
        if init_pos is not None:
            self.set(init_pos)
        self._egu = egu
        if limits is None:
            limits = (0, 0)

        self._limits = tuple(limits)
        self.source = source

    @property
    def limits(self):
        return self._limits

    @property
    def egu(self):
        '''The engineering units (EGU) for positions'''
        return self._egu

    def _setup_move(self, position, status):
        '''Move requested to position

        This is a SoftPositioner method which allows customization of what
        happens when a motion request happens without re-implementing
        all of `move`.

        Parameters
        ----------
        position : any
            Position to move to (already verified by `check_value`)
        status : MoveStatus
            Status object created by PositionerBase.move()
        '''
        # A soft positioner immediately 'moves' to the target position when
        # requested.
        self._run_subs(sub_type=self.SUB_START, timestamp=time.time())

        self._started_moving = True
        self._moving = False

        self._set_position(position)
        self._done_moving()

    def move(self, position, wait=True, timeout=None, moved_cb=None):
        '''Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback
            must accept one keyword argument: 'obj' which will be set to
            this positioner instance.
        wait : bool, optional
            Wait until motion has completed
        timeout : float, optional
            Maximum time to wait for a motion

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        '''
        status = super().move(position, moved_cb=moved_cb, timeout=timeout)

        self._setup_move(position, status)

        if wait:
            try:
                status_wait(status)
            except RuntimeError:
                raise RuntimeError('Motion did not complete successfully')

        return status

    def _repr_info(self):
        yield from super()._repr_info()
        yield ('egu', self._egu)
        yield ('limits', self._limits)
        yield ('source', self.source)

    def read(self):
        d = OrderedDict()
        d[self.name] = {'value': self.position,
                        'timestamp': time.time()}
        return d

    def describe(self):
        """Return the description as a dictionary

        Returns
        -------
        dict
            Dictionary of name and formatted description string
        """
        desc = OrderedDict()
        desc[self.name] = {'source': str(self.source),
                           'dtype': data_type(self.position),
                           'shape': data_shape(self.position),
                           'units': self.egu,
                           'lower_ctrl_limit': self.low_limit,
                           'upper_ctrl_limit': self.high_limit,
                           }
        return desc

    def read_configuration(self):
        return OrderedDict()

    def describe_configuration(self):
        return OrderedDict()



import logging
import pytest

import numpy as np

from ophyd import (Device, Component, FormattedComponent)
from ophyd.signal import (Signal, AttributeSignal, ArrayAttributeSignal)
from ophyd.utils import ExceptionBundle
from .conftest import AssertTools

logger = logging.getLogger(__name__)


class FakeSignal(Signal):
    def __init__(self, read_pv, *, name=None, parent=None, **kwargs):
        self.read_pv = read_pv
        super().__init__(name=name, parent=parent, **kwargs)

    def get(self):
        return self.name

    def describe_configuration(self):
        return {self.name + '_conf': {'source': 'SIM:test'}}

    def read_configuration(self):
        return {self.name + '_conf': {'value': 0}}


def setUpModule():
    pass


def tearDownModule():
    logger.debug('Cleaning up')


def test_device_state():
    d = Device('test', name='test')

    d.stage()
    old, new = d.configure({})
    assert old == new
    d.unstage()


class TestDevice(AssertTools):
    def test_attrs(self):
        class MyDevice(Device):
            cpt1 = Component(FakeSignal, '1')
            cpt2 = Component(FakeSignal, '2')
            cpt3 = Component(FakeSignal, '3')

        d = MyDevice('prefix', read_attrs=['cpt1'],
                     configuration_attrs=['cpt2'],
                     name='test'
                     )

        d.read()
        self.assertEqual(d.read_attrs, ['cpt1'])
        self.assertEqual(d.configuration_attrs, ['cpt2'])

        self.assertEqual(list(d.read().keys()), [d.cpt1.name])
        self.assertEqual(set(d.read_configuration().keys()),
                         {d.cpt2.name + '_conf'})

        self.assertEqual(list(d.describe().keys()), [d.cpt1.name])
        self.assertEqual(set(d.describe_configuration().keys()),
                         {d.cpt2.name + '_conf', })

    def test_complexdevice(self):
        class SubDevice(Device):
            cpt1 = Component(FakeSignal, '1')
            cpt2 = Component(FakeSignal, '2')
            cpt3 = Component(FakeSignal, '3')

        class SubSubDevice(SubDevice):
            cpt4 = Component(FakeSignal, '4')

        class MyDevice(Device):
            sub1 = Component(SubDevice, '1')
            subsub2 = Component(SubSubDevice, '2')
            cpt3 = Component(FakeSignal, '3')

        device = MyDevice('prefix', name='dev')
        device.configuration_attrs = ['sub1',
                                      'subsub2.cpt2',
                                      'subsub2.cpt4',
                                      'cpt3']
        device.sub1.read_attrs = ['cpt2']
        device.sub1.configuration_attrs = ['cpt1']

        self.assertIs(device.sub1.parent, device)
        self.assertIs(device.subsub2.parent, device)
        self.assertIs(device.cpt3.parent, device)

        self.assertEquals(device.sub1.component_names,
                          ('cpt1', 'cpt2', 'cpt3'))
        self.assertEquals(device.subsub2.component_names,
                          ('cpt1', 'cpt2', 'cpt3', 'cpt4'))

        with pytest.raises(AttributeError):
            device.component_names.remove('sub1')

        conf_keys = {'dev_sub1_cpt1_conf',    # from sub1.*
                     # 'dev_sub1_cpt2_conf',  # not in sub1.config_attrs
                     'dev_subsub2_cpt2_conf', # from subsub2.cpt2
                     'dev_subsub2_cpt4_conf', # from subsub2.cpt4
                     'dev_cpt3_conf',         # from cpt3


                     }

        self.assertEquals(set(device.describe_configuration().keys()),
                          conf_keys)
        self.assertEquals(set(device.read_configuration().keys()),
                          conf_keys)

    def test_complexdevice_stop(self):
        class SubSubDevice(Device):
            cpt4 = Component(FakeSignal, '4')

            def stop(self, *, success=False):
                self.stop_called = True
                self.success = success
                if self.prefix.endswith('_raises_'):
                    raise Exception('stop failed for some reason')

        class SubDevice(Device):
            cpt1 = Component(FakeSignal, '1')
            cpt2 = Component(FakeSignal, '2')
            cpt3 = Component(FakeSignal, '3')
            subsub = Component(SubSubDevice, '')

            def stop(self, *, success=False):
                self.stop_called = True
                self.success = success
                super().stop(success=success)

        class MyDevice(Device):
            sub1 = Component(SubDevice, '1')
            sub2 = Component(SubDevice, '_raises_')
            sub3 = Component(SubDevice, '_raises_')
            cpt3 = Component(FakeSignal, '3')

        dev = MyDevice('', name='mydev')
        with pytest.raises(ExceptionBundle) as cm:
            dev.stop()

        ex = cm.value
        self.assertEquals(len(ex.exceptions), 2)
        self.assertTrue(dev.sub1.stop_called)
        self.assertTrue(dev.sub2.stop_called)
        self.assertTrue(dev.sub3.stop_called)
        self.assertFalse(dev.sub1.success)
        self.assertFalse(dev.sub2.success)
        self.assertFalse(dev.sub3.success)

        self.assertTrue(dev.sub1.subsub.stop_called)
        self.assertTrue(dev.sub2.subsub.stop_called)
        self.assertTrue(dev.sub3.subsub.stop_called)
        self.assertFalse(dev.sub1.subsub.success)
        self.assertFalse(dev.sub2.subsub.success)
        self.assertFalse(dev.sub3.subsub.success)

        dev = MyDevice('', name='mydev')
        with pytest.raises(ExceptionBundle) as cm:
            dev.stop(success=True)

        ex = cm.value
        self.assertEquals(len(ex.exceptions), 2)
        self.assertTrue(dev.sub1.stop_called)
        self.assertTrue(dev.sub2.stop_called)
        self.assertTrue(dev.sub3.stop_called)
        self.assertTrue(dev.sub1.success)
        self.assertTrue(dev.sub2.success)
        self.assertTrue(dev.sub3.success)

        self.assertTrue(dev.sub1.subsub.stop_called)
        self.assertTrue(dev.sub2.subsub.stop_called)
        self.assertTrue(dev.sub3.subsub.stop_called)
        self.assertTrue(dev.sub1.subsub.success)
        self.assertTrue(dev.sub2.subsub.success)
        self.assertTrue(dev.sub3.subsub.success)

    def test_name_shadowing(self):
        RESERVED_ATTRS = ['name', 'parent', 'component_names', '_signals',
                          '_sig_attrs',
                          '_sub_devices']

        type('a', (Device,), {'a': None})  # legal class definition
        # Illegal class definitions:
        for attr in RESERVED_ATTRS:
            self.assertRaises(TypeError, type, 'a', (Device,), {attr: None})

    def test_formatted_component(self):
        FC = FormattedComponent

        class MyDevice(Device):
            cpt = Component(FakeSignal, 'suffix')
            ch = FC(FakeSignal, '{self.prefix}{self._ch}')

            def __init__(self, prefix, ch='a', **kwargs):
                self._ch = ch
                super().__init__(prefix, **kwargs)

        ch_value = '_test_'

        device = MyDevice('prefix:', ch=ch_value, name='test')
        self.assertIs(device.cpt.parent, device)
        self.assertIs(device.ch.parent, device)
        self.assertIs(device._ch, ch_value)
        self.assertEquals(device.ch.read_pv, device.prefix + ch_value)
        self.assertEquals(device.cpt.read_pv,
                          device.prefix + MyDevice.cpt.suffix)

    def test_root(self):
        class MyDevice(Device):
            cpt = Component(FakeSignal, 'suffix')

        d = MyDevice('', name='test')
        assert d.cpt.root == d
        assert d.root == d

    def test_hidden_component(self):
        class MyDevice(Device):
            _hidden_sig = Component(FakeSignal, 'suffix')
        d = MyDevice('', name='test')
        assert '_hidden_sig' in d.component_names
        assert not hasattr(d.get(), '_hidden_sig')


def test_attribute_signal():
    init_value = 33

    class SubDevice(Device):
        @property
        def prop(self):
            return init_value

        @prop.setter
        def prop(self, value):
            pass

    class MyDevice(Device):
        sub1 = Component(SubDevice, '1')
        attrsig = Component(AttributeSignal, 'prop')
        sub_attrsig = Component(AttributeSignal, 'sub1.prop')

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._value = init_value

        @property
        def prop(self):
            return self._value

        @prop.setter
        def prop(self, value):
            self._value = value

    dev = MyDevice('', name='mydev')
    assert (dev.describe()['mydev_attrsig']['source'] == 'PY:mydev.prop')
    assert (dev.describe()['mydev_sub_attrsig']['source'] == 'PY:mydev.sub1.prop')
    assert dev.attrsig.get() == init_value
    dev.attrsig.put(55)
    assert dev.attrsig.get() == 55

    assert dev.sub_attrsig.get() == init_value
    dev.sub_attrsig.put(0)


def test_attribute_signal_attributeerror():
    class MyDevice(Device):
        sig1 = Component(AttributeSignal, 'unknownattribute')
        sig2 = Component(AttributeSignal, '__class__.__name')
        sig3 = Component(AttributeSignal, 'unknown.attribute')

    dev = MyDevice('', name='mydev')
    pytest.raises(AttributeError, dev.sig1.get)
    pytest.raises(AttributeError, dev.sig2.get)
    pytest.raises(AttributeError, dev.sig3.get)
    pytest.raises(AttributeError, getattr, dev.sig3, 'base')


def test_shadowing_bs_interface_raises_typeerror():
    class LegalDevice(Device):
        cpt = Component(FakeSignal, 'cpt')

    with pytest.raises(TypeError):
        class IllegalDevice(Device):
            # 'read' shadows the bluesky interface and should not be allowed
            read = Component(FakeSignal, 'read')


def test_array_attribute_signal():
    init_value = range(10)

    class MyDevice(Device):
        attrsig = Component(ArrayAttributeSignal, 'prop')

        @property
        def prop(self):
            return init_value

    dev = MyDevice('', name='mydev')
    np.testing.assert_array_equal(dev.attrsig.get(), init_value)
    assert isinstance(dev.attrsig.get(), np.ndarray)
    assert isinstance(dev.attrsig.get(), np.ndarray)


def test_signal_names():

    class MyDevice(Device):
        cpt = Component(FakeSignal, 'suffix')

    d = MyDevice('', name='test')
    with pytest.warns(UserWarning):
        signal_names = d.signal_names

    assert signal_names is d.component_names


def test_summary():
    class MyDevice(Device):
        cpt = Component(FakeSignal, 'suffix')

    d = MyDevice('', name='test')
    # smoke tests
    d.summary()
    d.__str__()
    d.__repr__()

def test_labels():
    class MyDevice(Device):
        cpt = Component(FakeSignal, 'suffix')

    d = MyDevice('', name='test', labels={'a', 'b'})
    assert d._ophyd_labels_ == {'a', 'b'}



from ophyd import (Device, Signal, Kind, Component, ALL_COMPONENTS,
                   kind_context, DynamicDeviceComponent as DDC)
import pytest


def test_standalone_signals():
    # A object's kind only matters when we read its _parent_. It affects
    # whether its parent recursively calls read() and/or read_configuration().
    # When we call read() and/or read_configuration() on the object itself,
    # directly, and its 'kind' setting does not affects its behavior.
    normal_sig = Signal(name='normal_sig', value=3, kind=Kind.normal)
    assert normal_sig.kind == Kind.normal
    assert normal_sig.name not in normal_sig.hints['fields']
    assert 'normal_sig' in normal_sig.read()
    assert 'normal_sig' in normal_sig.read_configuration()

    hinted_sig = Signal(name='hinted_sig', value=3)
    assert hinted_sig.kind == Kind.hinted
    assert hinted_sig.name in hinted_sig.hints['fields']
    assert 'hinted_sig' in hinted_sig.read()
    assert 'hinted_sig' in hinted_sig.read_configuration()

    # Same with a sig set up this way
    config_sig = Signal(name='config_sig', value=5, kind=Kind.config)
    assert config_sig.kind == Kind.config
    assert config_sig.name not in config_sig.hints['fields']
    assert 'config_sig' in config_sig.read()
    assert 'config_sig' in config_sig.read_configuration()

    # Same with a sig set up this way
    omitted_sig = Signal(name='omitted_sig', value=5, kind=Kind.omitted)
    assert omitted_sig.kind == Kind.omitted
    assert omitted_sig.name not in omitted_sig.hints['fields']
    assert 'omitted_sig' in omitted_sig.read()
    assert 'omitted_sig' in omitted_sig.read_configuration()

    # But these will be differentiated when we put them into a parent
    # Device and call read() or read_configuraiton() on _that_.


def test_nested_devices():

    class A(Device):
        normal_sig = Component(Signal)
        config_sig = Component(Signal, kind=Kind.config)
        omitted_sig = Component(Signal, kind=Kind.omitted)

    a = A(name='a')

    # When we call read and read_configuration on a, it checks the kind of its
    # components and reads the right ones.
    assert 'a_normal_sig' in a.read()
    assert 'a_config_sig' not in a.read()
    assert 'a_omitted_sig' not in a.read()

    assert 'a_normal_sig' not in a.read_configuration()
    assert 'a_config_sig' in a.read_configuration()
    assert 'a_omitted_sig' not in a.read_configuration()

    # Another layer of nesting!

    class B(Device):
        a_default = Component(A)
        a_config = Component(A, kind=Kind.config)
        a_omitted = Component(A, kind=Kind.omitted)

    b = B(name='b')

    assert ['b_a_default_normal_sig'] == list(b.read())

    # Notice that a_default comes along for the ride here. If you ask
    # a Device for its normal readings it will also give you its
    # configuration. (You need complete configurational metadata for
    # the EventDescriptor!)
    assert ['b_a_default_config_sig',
            'b_a_config_config_sig'] == list(b.read_configuration())
    # Notice that it tacks 'config' on when you set the kind to 'normal'.
    assert (Kind.normal | Kind.config) == B(name='b', kind=Kind.normal).kind
    # And the same if you try to set it after __init__
    b.a_default.kind = Kind.normal
    assert b.a_default.kind == (Kind.normal | Kind.config)
    # However, just taking 'config' alone without Event-wise readings is fine.
    b.a_default.kind = Kind.config
    assert b.a_default.kind == Kind.config
    # Now we get no Event-wise readings from a_default
    assert [] == list(b.read())
    assert ['b_a_default_config_sig',
            'b_a_config_config_sig'] == list(b.read_configuration())


def test_strings():
    sig = Signal(name='sig', kind='normal')
    assert sig.kind == Kind.normal


def test_kind_context():
    class Thing(Device):

        with kind_context('omitted') as Cpt:
            a = Cpt(Signal)

    thing = Thing(name='thing')
    assert thing.a.kind == Kind.omitted


# Test back-compatibility. The expected values in these tests match the
# behaviors from before the Kind feature was introduced --- modulo the _order_.


def test_behavior_if_nothing_is_specified():

    class Thing(Device):
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name='t')
    assert set('ab') == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)


def test_class_default_attrs_both_none():

    # Test various values of _default_read_attrs and
    # _default_configuration_attrs

    class Thing(Device):
        _default_read_attrs = None
        _default_configuration_attrs = None
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name='t')
    assert set('ab') == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)
    assert ['t_a', 't_b'] == list(t.describe())
    assert set() == set(t.describe_configuration())

    class ThingHaver(Device):
        _default_read_attrs = None
        _default_configuration_attrs = None
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name='th')
    assert set(['a', 'b'] + ['a.a', 'a.b', 'b.a', 'b.b']) == set(th.read_attrs)

    # THIS IS ACTUALLY _NOT_ BACKWARD-COMPATIBLE BEHAVIOR!
    # But we consider the former behavior a bug. Unlike a Signal, a Device is
    # not allowed to place itself in its parent's read_attrs ONLY. It must also
    # be in configuration_attrs if it is in read_attrs. (Under the hood, this
    # is enforced in the setter of the Device's `kind` property, which OR's the
    # kind value with Kind.config if said vlaue includes Kind.normal.
    assert set('ab') == set(th.configuration_attrs)

    assert ['th_a_a', 'th_a_b', 'th_b_a', 'th_b_b'] == list(th.describe())
    assert [] == list(t.describe_configuration())


def test_default_read_attrs_empty_and_configuration_attrs_none():

    class Thing(Device):
        _default_read_attrs = []
        _default_configuration_attrs = None
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name='t')
    assert set() == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)

    class ThingHaver(Device):
        _default_read_attrs = []
        _default_configuration_attrs = None
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name='th')
    assert set() == set(th.read_attrs)
    assert [] == list(th.describe())
    assert set() == set(th.configuration_attrs)
    assert [] == list(th.describe_configuration())


def test_default_read_attrs_none_and_configuration_attrs_empty():

    class Thing(Device):
        _default_read_attrs = None
        _default_configuration_attrs = []
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name='t')
    assert set('ab') == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)

    class ThingHaver(Device):
        _default_read_attrs = None
        _default_configuration_attrs = []
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name='th')
    assert set(['a', 'b'] + ['a.a', 'a.b', 'b.a', 'b.b']) == set(th.read_attrs)
    assert ['th_a_a', 'th_a_b', 'th_b_a', 'th_b_b'] == list(th.describe())
    assert set('ab') == set(th.configuration_attrs)
    assert [] == list(th.describe_configuration())


def test_default_attrs_both_empty():

    class Thing(Device):
        _default_read_attrs = []
        _default_configuration_attrs = []
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name='t')
    assert set() == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)

    class ThingHaver(Device):
        _default_read_attrs = []
        _default_configuration_attrs = []
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name='th')
    assert set() == set(th.read_attrs)
    assert [] == list(th.describe())
    assert set() == set(th.configuration_attrs)
    assert [] == list(t.describe_configuration())


def test_all_components_escape_hatch():

    class Thing(Device):
        _default_read_attrs = ['a']
        _default_configuration_attrs = ['b']
        a = Component(Signal)
        b = Component(Signal)

    class ThingEscapeHatch(Thing):
        _default_read_attrs = ALL_COMPONENTS
        _default_configuration_attrs = []
        c = Component(Signal)

    t = ThingEscapeHatch(name='t')
    assert set('abc') == set(t.read_attrs)
    assert set() == set(t.configuration_attrs)


def test_default_attrs_nonempty_disjoint():

    class Thing(Device):
        _default_read_attrs = ['a']
        _default_configuration_attrs = ['b']
        a = Component(Signal)
        b = Component(Signal)

    t = Thing(name='t')
    assert set('a') == set(t.read_attrs)
    assert set('b') == set(t.configuration_attrs)

    class ThingHaver(Device):
        _default_read_attrs = ['a']
        _default_configuration_attrs = ['b']
        a = Component(Thing)
        b = Component(Thing)

    th = ThingHaver(name='th')
    assert set(['a', 'a.a']) == set(th.read_attrs)
    assert 'b' not in th.read_attrs
    assert ['th_a_a'] == list(th.describe())
    # Here again we have the same break from back-compatability. We are not
    # allowed to put a sub-Device in configuration_attrs that is not also in
    # read_attrs. The old code would have:
    # assert set(['b']) == set(th.configuration_attrs)
    # But now we get:
    assert set(['a', 'b'] + ['a.b', 'b.b']) == set(th.configuration_attrs)
    assert ['th_a_b', 'th_b_b'] == list(th.describe_configuration())


def test_options_via_init():
    ...


def test_back_compat():
    # Test class defaults.
    class Thing(Device):
        _default_read_attrs = ['a']
        _default_configuration_attrs = ['b']
        a = Component(Signal)
        b = Component(Signal)
        c = Component(Signal)

    thing = Thing(name='thing')
    assert ['thing_a'] == list(thing.read())
    # assert ['thing_b'] == list(thing.read_configuration())

    # Test attribute getting and setting.
    assert thing.read_attrs == ['a']
    assert thing.configuration_attrs == ['b']
    thing.read_attrs = ['a', 'b']
    assert ['thing_a', 'thing_b'] == list(thing.read())

    thing = Thing(name='thing', read_attrs=['b'], configuration_attrs=['a'])
    assert ['thing_b'] == list(thing.read())


@pytest.fixture(scope='function')
def thing_haver_haver():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal)
        d = Component(Signal, kind=Kind.hinted)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted)
        B = Component(Thing, kind=Kind.config)
        C = Component(Thing, kind=Kind.normal)

    class ThingHaverHaver(Device):
        alpha = Component(ThingHaver, kind=Kind.omitted)
        beta = Component(ThingHaver, kind=Kind.config)
        gamma = Component(ThingHaver, kind=Kind.normal)

    return ThingHaverHaver(name='thh')


def test_list_proxy(thing_haver_haver):

    thh = thing_haver_haver
    assert 'gamma' in thh.read_attrs
    assert 'gamma' in list(thh.read_attrs)

    assert 'beta' in thh.configuration_attrs
    assert 'beta' in list(thh.configuration_attrs)

    assert 'alpha' not in thh.read_attrs
    assert 'alpha' not in thh.configuration_attrs

    assert 'alpha' not in list(thh.read_attrs)
    assert 'alpha' not in list(thh.configuration_attrs)

    assert 'gamma.C' in thh.read_attrs
    assert 'gamma.C' in list(thh.read_attrs)

    assert 'gamma.C.c' in list(thh.read_attrs)
    assert 'gamma.C.d' in list(thh.read_attrs)


def test_ddc():

    class Thing(Device):
        a = Component(Signal)
        b = Component(Signal)

    class ThingHaver(Device):
        a = DDC({'A': (Thing, '', {})})


    th = ThingHaver(name='th')
    assert th.a.A.a.kind & Kind.normal
    assert th.a.A.kind & Kind.normal
    assert th.a.kind & Kind.normal
    assert th.kind & Kind.normal


def test_no_lazy():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal)
        d = Component(Signal, kind=Kind.hinted)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted)
        B = Component(Thing, kind=Kind.config)
        C = Component(Thing, kind=Kind.normal)

    th = ThingHaver(name='th')
    assert set('ABC') == set(th._signals)
    for k in 'ABC':
        assert k in th._signals
        assert set('abcd') == set(getattr(th, k)._signals)


def test_some_lazy():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal, lazy=True)
        d = Component(Signal, kind=Kind.hinted, lazy=True)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted, lazy=True)
        B = Component(Thing, kind=Kind.config)
        C = Component(Thing, kind=Kind.normal)

    th = ThingHaver(name='th')
    assert set('BC') == set(th._signals)
    for k in 'BC':
        c = getattr(th, k)
        assert set('ab') == set(c._signals)


def test_all_lazy():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal, lazy=True)
        d = Component(Signal, kind=Kind.hinted, lazy=True)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted, lazy=True)
        B = Component(Thing, kind=Kind.config, lazy=True)
        C = Component(Thing, kind=Kind.normal, lazy=True)

    th = ThingHaver(name='th')
    assert not th._signals


def test_grand_child_breaks_lazy():
    class Thing(Device):
        a = Component(Signal, kind=Kind.omitted)
        b = Component(Signal, kind=Kind.config)
        c = Component(Signal, kind=Kind.normal, lazy=True)
        d = Component(Signal, kind=Kind.hinted, lazy=True)

    class ThingHaver(Device):
        A = Component(Thing, kind=Kind.omitted, lazy=True)
        B = Component(Thing, kind=Kind.config, lazy=True)
        C = Component(Thing, kind=Kind.normal, lazy=True)

    th = ThingHaver(name='th', read_attrs=['A.a', 'A.d'])
    assert set('A') == set(th._signals)
    A = th.A
    assert set('ab') == set(A._signals)



import textwrap
import inspect
import re
import sys
from collections import OrderedDict
import networkx as nx

from ..signal import EpicsSignal
from . import docs
from ..device import (Device, Component)
from ..signal import (ArrayAttributeSignal)


class EpicsSignalWithRBV(EpicsSignal):
    # An EPICS signal that simply uses the areaDetector convention of
    # 'pvname' being the setpoint and 'pvname_RBV' being the read-back

    def __init__(self, prefix, **kwargs):
        super().__init__(prefix + '_RBV', write_pv=prefix, **kwargs)


class ADComponent(Component):
    def __init__(self, cls, suffix, **kwargs):
        super().__init__(cls, suffix, lazy=True, **kwargs)

    def find_docs(self, parent_class):
        '''Find all the documentation related to this class, all the way up the
        MRO'''

        classes = inspect.getmro(parent_class)
        for class_ in classes:
            try:
                html_file = class_._html_docs
            except AttributeError:
                continue

            for fn in html_file:
                if fn in docs.docs:
                    yield docs.docs[fn]

    def make_docstring(self, parent_class):
        '''Create a docstring for the component, given the parent class'''
        def make_codeblock(s):
            '''Make a codeblock that will render nicely in sphinx'''
            block = ['AreaDetector Component',
                     '::',
                     '',
                     ]

            lines = s.split('\n', 1)
            header, lines = lines[0], lines[1:]

            block.append(textwrap.indent(textwrap.dedent(header),
                                         prefix=' ' * 4))

            lines = '\n'.join(lines)
            block.append(textwrap.indent(textwrap.dedent(lines),
                                         prefix=' ' * 4))
            block.append('')
            return '\n'.join(block)

        suffixes = [self.suffix]

        if self.suffix.endswith('_RBV'):
            suffixes.append(self.suffix[:-4])

        for doc in self.find_docs(parent_class):
            for suffix in suffixes:
                try:
                    return make_codeblock(doc[suffix])
                except KeyError:
                    pass

        return super().make_docstring(parent_class)


def ad_group(cls, attr_suffix, **kwargs):
    '''Definition creation for groups of signals in areadetectors'''
    defn = OrderedDict()
    kwargs.setdefault('lazy', True)
    for attr, suffix in attr_suffix:
        defn[attr] = (cls, suffix, kwargs)
    return defn


class ADBase(Device):
    '''The AreaDetector base class

    This serves as the base for all detectors and plugins
    '''

    _html_docs = ['areaDetectorDoc.html']
    _default_read_attrs = ()
    _default_configuration_attrs = ()

    def find_signal(self, text, use_re=False, case_sensitive=False,
                    match_fcn=None, f=sys.stdout):
        '''Search through the signal docs on this detector for the string text

        Parameters
        ----------
        text : str
            Text to find
        use_re : bool, optional
            Use regular expressions
        case_sensitive : bool, optional
            Case sensitive search
        match_fcn : callable, optional
            Function to call when matches are found Defaults to a function that
            prints matches to f
        f : file-like, optional
            File-like object that the default match function prints to
            (Defaults to sys.stdout)
        '''
        # TODO: Some docstrings change based on the detector type,
        #       showing different options than are available in
        #       the base area detector class (for example). As such,
        #       instead of using the current docstrings, this grabs
        #       them again.

        def default_match(attr, signal, doc):
            print('Property: {}'.format(attr), file=f)
            print('  Signal: {!r}'.format(signal), file=f)
            print('     Doc: {}'.format(doc), file=f)
            print(file=f)

        if match_fcn is None:
            match_fcn = default_match

        if use_re:
            flags = re.MULTILINE
            if not case_sensitive:
                flags |= re.IGNORECASE

            regex = re.compile(text, flags=flags)

        elif not case_sensitive:
            text = text.lower()

        for attr, cpt in self._sig_attrs.items():
            doc = cpt.make_docstring(self.__class__)

            match = False
            if use_re:
                if regex.search(doc):
                    match = True
            else:
                if not case_sensitive:
                    if text in doc.lower():
                        match = True
                elif text in doc:
                    match = True

            if match:
                match_fcn(attr=attr, signal=getattr(self, attr), doc=doc)

    def stage(self, *args, **kwargs):
        ret = super().stage(*args, **kwargs)
        try:
            self.validate_asyn_ports()
        except RuntimeError as err:
            self.unstage(*args, **kwargs)
            raise err
        return ret

    def get_plugin_by_asyn_port(self, port_name):
        '''Get the plugin which has the given asyn port name

        Parameters
        ----------
        port_name : str
            The port name to search for

        Returns
        -------
        ret : ADBase or None
            Either the requested plugin or None if not found

        '''
        try:
            name = self.port_name.get()
        except AttributeError:
            pass
        else:
            if name == port_name:
                return self

        for nm in self._sub_devices:
            cpt = getattr(self, nm)
            if hasattr(cpt, 'get_plugin_by_asyn_port'):
                sig = cpt.get_plugin_by_asyn_port(port_name)
                if sig is not None:
                    return sig
        return None

    def get_asyn_port_dictionary(self):
        '''Return port name : component map

        Returns
        -------
        port_map : dict
            Mapping between port_name and ADBase objects
        '''
        # uniqueness of port names enforced at IOC layer
        ret = {}
        try:
            ret.update({self.port_name.get(): self})
        except AttributeError:
            pass

        for nm in self._sub_devices:
            sig = getattr(self, nm)
            if hasattr(sig, 'get_asyn_port_dictionary'):
                ret.update(sig.get_asyn_port_dictionary())

        return ret

    def get_asyn_digraph(self):
        '''Get the directed graph of the ASYN ports

        Returns
        -------
        G : networkx.DiGraph
            Directed graph of pipelines

        port_map : dict
            Mapping between port_name and ADBase objects
        '''
        port_map = self.get_asyn_port_dictionary()
        G = nx.DiGraph()
        for out_port, cpt in port_map.items():
            try:
                in_port = cpt.nd_array_port.get()
            except AttributeError:
                # attribute error because we hit a component which is not
                # a plugin, but is the 'base' data source
                G.add_node(out_port)
            else:
                G.add_edge(in_port, out_port)

        return G, port_map

    def visualize_asyn_digraph(self, ax=None, *args, **kwargs):
        '''This generates a figure showing the current asyn port layout.

        This method generates a plot showing all of the currently enabled
        Areadetector plugin asyn ports and their relationships. The current
        ports and relationships are found using self.get_asyn_digraph.

        Parameters
        ----------
        ax: matplotlib axes
            if None (default) then a new figure is created otherwise it is
            plotted on the specified axes.
        *args, **kwargs : networkx.draw_networkx args and kwargs.
            For the allowed args and kwargs see the `networkx.draw_networkx documentation <https://networkx.github.io/documentation/networkx-1.10/reference/generated/networkx.drawing.nx_pylab.draw_networkx.html>`_
        '''
        # Importing matplotlib.pyplot here as it is not a dependency except for
        # this method.
        import matplotlib.pyplot as plt

        # Generate the port_map Digraph.
        G, port_map = self.get_asyn_digraph()
        # Create and label the figure if no ax is provided.
        if not ax:
            fig, ax = plt.subplots()
            ax.set_title('AD port map for {}'.format(self.name))
            plt.tick_params(axis='x', which='both', bottom=False, top=False,
                            labelbottom=False)
            plt.tick_params(axis='y', which='both', left=False, right=False,
                            labelbottom=False)

        nx.draw_networkx(G, ax=ax, *args, **kwargs)

    def validate_asyn_ports(self):
        '''Validate that all components of pipeline are known

        Raises
        ------
        RuntimeError
           If there any input ports to known plugins where the source is
           not known to ophyd
        '''
        g, port_map = self.get_asyn_digraph()
        g = nx.Graph(g)
        if port_map and nx.number_connected_components(g) != 1:
            missing_plugins = self.missing_plugins()
            raise RuntimeError('The asyn ports {!r} are used by plugins '
                               'that ophyd is aware of but the source plugin '
                               'is not.  Please reconfigure your device to '
                               'include the source plugin or reconfigure '
                               'to not use these ports.'
                               ''.format(missing_plugins))

    def missing_plugins(self):
        '''Find missing ports'''
        g, port_map = self.get_asyn_digraph()
        ret = []

        for node in g:
            if node not in port_map:
                ret.append(node)

        return ret

    configuration_names = Component(ArrayAttributeSignal,
                                    attr='_configuration_names')

    @property
    def _configuration_names(self):
        return [getattr(self, c).name
                for c in self.configuration_attrs]

    @property
    def ad_root(self):
        root = self
        while True:
            if not isinstance(root.parent, ADBase):
                return root
            root = root.parent



import logging
import time
import copy
import pytest
from ophyd import get_cl

from ophyd.signal import (Signal, EpicsSignal, EpicsSignalRO, DerivedSignal)
from ophyd.utils import ReadOnlyError
from ophyd.status import wait

from .conftest import (FakeEpicsWaveform, using_fake_epics_pv,
                       using_fake_epics_waveform)
logger = logging.getLogger(__name__)


@using_fake_epics_pv
def test_fakepv():
    pvname = 'fakepv_nowaythisexists' * 10

    info = dict(called=False)

    def conn(**kwargs):
        info['conn'] = True
        info['conn_kw'] = kwargs

    def value_cb(**kwargs):
        info['value'] = True
        info['value_kw'] = kwargs
    cl = get_cl()
    pv = cl.get_pv(pvname, callback=value_cb, connection_callback=conn)

    if not pv.wait_for_connection():
        raise ValueError('should return True on connection')

    assert pv.pvname == pvname

    pv._update_rate = 0.5
    time.sleep(0.2)

    assert info['conn']
    assert info['value']
    assert info['value_kw']['value'] == pv.value


@using_fake_epics_pv
def test_fakepv_signal():
    sig = EpicsSignal(write_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.VAL',
                      read_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.RBV')
    st = sig.set(1)

    for j in range(10):
        if st.done:
            break
        time.sleep(.1)

    assert st.done


def test_signal_base():
    start_t = time.time()

    name = 'test'
    value = 10.0
    signal = Signal(name=name, value=value, timestamp=start_t)
    signal.wait_for_connection()

    assert signal.connected
    assert signal.name == name
    assert signal.value == value
    assert signal.get() == value
    assert signal.timestamp == start_t

    info = dict(called=False)

    def _sub_test(**kwargs):
        info['called'] = True
        info['kw'] = kwargs

    signal.subscribe(_sub_test, run=False,
                     event_type=signal.SUB_VALUE)
    assert not info['called']

    signal.value = value
    signal.clear_sub(_sub_test)

    signal.subscribe(_sub_test, run=False,
                     event_type=signal.SUB_VALUE)
    signal.clear_sub(_sub_test, event_type=signal.SUB_VALUE)

    kw = info['kw']
    assert 'value' in kw
    assert 'timestamp' in kw
    assert 'old_value' in kw

    assert kw['value'] == value
    assert kw['old_value'] == value
    assert kw['timestamp'] == signal.timestamp

    # readback callback for soft signal
    info = dict(called=False)
    signal.subscribe(_sub_test, event_type=Signal.SUB_VALUE,
                     run=False)
    assert not info['called']
    signal.put(value + 1)
    assert info['called']

    signal.clear_sub(_sub_test)
    kw = info['kw']

    assert 'value' in kw
    assert 'timestamp' in kw
    assert 'old_value' in kw

    assert kw['value'] == value + 1
    assert kw['old_value'] == value
    assert kw['timestamp'] == signal.timestamp

    signal.trigger()
    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))


def test_signal_copy():
    start_t = time.time()

    name = 'test'
    value = 10.0
    signal = Signal(name=name, value=value, timestamp=start_t)
    sig_copy = copy.copy(signal)

    assert signal.name == sig_copy.name
    assert signal.value == sig_copy.value
    assert signal.get() == sig_copy.get()
    assert signal.timestamp == sig_copy.timestamp


def test_rw_removal():
    # rw kwarg is no longer used
    with pytest.raises(RuntimeError):
        EpicsSignal('readpv', rw=False)

    with pytest.raises(RuntimeError):
        EpicsSignal('readpv', rw=True)


@using_fake_epics_pv
def test_epicssignal_readonly():
    signal = EpicsSignalRO('readpv')
    signal.wait_for_connection()
    signal.value

    with pytest.raises(ReadOnlyError):
        signal.value = 10

    with pytest.raises(ReadOnlyError):
        signal.put(10)

    with pytest.raises(ReadOnlyError):
        signal.set(10)

    # vestigial, to be removed
    with pytest.raises(AttributeError):
        signal.setpoint_ts

    # vestigial, to be removed
    with pytest.raises(AttributeError):
        signal.setpoint

    signal.precision
    signal.timestamp
    signal.limits

    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))
    time.sleep(0.2)


@using_fake_epics_pv
def test_epicssignal_readwrite_limits():
    signal = EpicsSignal('readpv', write_pv='readpv', limits=True)

    signal.wait_for_connection()
    signal.check_value((signal.low_limit + signal.high_limit) / 2)

    try:
        signal.check_value(None)
    except ValueError:
        pass
    else:
        raise ValueError('value=None')

    try:
        signal.check_value(signal.low_limit - 1)
    except ValueError:
        pass
    else:
        raise ValueError('lower limit %s' % (signal.limits, ))

    try:
        signal.check_value(signal.high_limit + 1)
    except ValueError:
        pass
    else:
        raise ValueError('upper limit')


@using_fake_epics_pv
def test_epicssignal_readwrite():
    signal = EpicsSignal('readpv', write_pv='writepv')

    signal.wait_for_connection()
    assert signal.setpoint_pvname == 'writepv'
    assert signal.pvname == 'readpv'
    signal.value

    signal._update_rate = 2
    time.sleep(0.2)

    value = 10
    signal.value = value
    signal.setpoint = value
    assert signal.setpoint == value
    signal.setpoint_ts

    signal.limits
    signal.precision
    signal.timestamp

    signal.read()
    signal.describe()
    signal.read_configuration()
    signal.describe_configuration()

    eval(repr(signal))
    time.sleep(0.2)


@using_fake_epics_waveform
def test_epicssignal_waveform():
    def update_cb(value=None, **kwargs):
        assert value in FakeEpicsWaveform.strings

    signal = EpicsSignal('readpv', string=True)

    signal.wait_for_connection()

    signal.subscribe(update_cb, event_type=signal.SUB_VALUE)
    assert signal.value in FakeEpicsWaveform.strings


@using_fake_epics_pv
def test_no_connection():
    # special case in FakeEpicsPV that returns false in wait_for_connection
    sig = EpicsSignal('does_not_connect')
    pytest.raises(TimeoutError, sig.wait_for_connection)

    sig = EpicsSignal('does_not_connect')
    pytest.raises(TimeoutError, sig.put, 0.0)
    pytest.raises(TimeoutError, sig.get)

    sig = EpicsSignal('connects', write_pv='does_not_connect')
    pytest.raises(TimeoutError, sig.wait_for_connection)


@using_fake_epics_pv
def test_enum_strs():
    sig = EpicsSignal('connects')
    sig.wait_for_connection()

    enums = ['enum_strs']

    # hack this onto the FakeEpicsPV
    sig._read_pv.enum_strs = enums

    assert sig.enum_strs == enums


@using_fake_epics_pv
def test_setpoint():
    sig = EpicsSignal('connects')
    sig.wait_for_connection()

    sig.get_setpoint()
    sig.get_setpoint(as_string=True)


def test_epicssignalro():
    # not in initializer parameters anymore
    pytest.raises(TypeError, EpicsSignalRO, 'test',
                  write_pv='nope_sorry')


@using_fake_epics_pv
def test_describe():
    sig = EpicsSignal('my_pv')
    sig._write_pv.enum_strs = ('enum1', 'enum2')
    sig.wait_for_connection()

    sig.put(1)
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'integer'
    assert desc['shape'] == []
    assert 'precision' in desc
    assert 'enum_strs' in desc
    assert 'upper_ctrl_limit' in desc
    assert 'lower_ctrl_limit' in desc

    sig.put('foo')
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'string'
    assert desc['shape'] == []

    sig.put(3.14)
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'number'
    assert desc['shape'] == []

    import numpy as np
    sig.put(np.array([1,]))
    desc = sig.describe()['my_pv']
    assert desc['dtype'] == 'array'
    assert desc['shape'] == [1,]


def test_set_method():
    sig = Signal(name='sig')

    st = sig.set(28)
    wait(st)
    assert st.done
    assert st.success
    assert sig.get() == 28


def test_soft_derived():
    timestamp = 1.0
    value = 'q'
    original = Signal(name='original', timestamp=timestamp, value=value)

    cb_values = []

    def callback(value=None, **kwargs):
        nonlocal cb_values
        cb_values.append(value)

    derived = DerivedSignal(derived_from=original, name='derived')
    derived.subscribe(callback, event_type=derived.SUB_VALUE)

    assert derived.timestamp == timestamp
    assert derived.get() == value
    assert derived.timestamp == timestamp
    assert derived.describe()[derived.name]['derived_from'] == original.name

    new_value = 'r'
    derived.put(new_value)
    assert original.get() == new_value
    assert derived.get() == new_value
    assert derived.timestamp == original.timestamp
    assert derived.limits == original.limits

    copied = copy.copy(derived)
    assert copied.derived_from.value == original.value
    assert copied.derived_from.timestamp == original.timestamp
    assert copied.derived_from.name == original.name

    derived.put('s')
    assert cb_values == ['r', 's']


@using_fake_epics_pv
def test_epics_signal_derived():
    signal = EpicsSignalRO('fakepv', name='original')

    derived = DerivedSignal(derived_from=signal, name='derived')
    derived.wait_for_connection()

    derived.connected

    # race condition with the FakeEpicsPV update loop, can't really test
    # assert derived.timestamp == signal.timestamp
    # assert derived.get() == signal.value


@pytest.mark.parametrize('put_complete', [True, False])
def test_epicssignal_set(put_complete):
    sim_pv = EpicsSignal(write_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.VAL',
                         read_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.RBV',
                         put_complete=put_complete)
    sim_pv.wait_for_connection()

    logging.getLogger('ophyd.signal').setLevel(logging.DEBUG)
    logging.getLogger('ophyd.utils.epics_pvs').setLevel(logging.DEBUG)
    print('tolerance=', sim_pv.tolerance)
    assert sim_pv.tolerance is not None

    start_pos = sim_pv.get()

    # move to +0.2 and check the status object
    target = sim_pv.get() + 0.2
    st = sim_pv.set(target, timeout=1, settle_time=0.001)
    wait(st, timeout=5)
    assert st.done
    assert st.success
    print('status 1', st)
    assert abs(target - sim_pv.get()) < 0.05

    # move back to -0.2, forcing a timeout with a low value
    target = sim_pv.get() - 0.2
    st = sim_pv.set(target, timeout=1e-6)
    time.sleep(0.1)
    print('status 2', st)
    assert st.done
    assert not st.success

    # keep the axis in position
    st = sim_pv.set(start_pos)
    wait(st, timeout=5)



def test_epicssignal_alarm_status():
    sig = EpicsSignal(write_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.VAL',
                      read_pv='XF:31IDA-OP{Tbl-Ax:X1}Mtr.RBV')
    sig.alarm_status
    sig.alarm_severity
    sig.setpoint_alarm_status
    sig.setpoint_alarm_severity


def test_epicssignalro_alarm_status():
    sig = EpicsSignalRO('XF:31IDA-OP{Tbl-Ax:X1}Mtr.RBV')
    sig.alarm_status
    sig.alarm_severity


def test_hints():
    sig = EpicsSignalRO('XF:31IDA-OP{Tbl-Ax:X1}Mtr.RBV')
    assert sig.hints == {'fields': [sig.name]}



import atexit
import ctypes
import epics
import queue
import threading
import warnings

from epics import get_pv as _get_pv, caget, caget, caput

try:
    epics.ca.find_libca()
except epics.ca.ChannelAccessException:
    thread_class = threading.Thread
else:
    thread_class = epics.ca.CAThread


_dispatcher = None


def get_pv(*args, **kwargs):
    import epics
    kwargs.setdefault('context', epics.ca.current_context())
    return _get_pv(*args, **kwargs)


class MonitorDispatcher(epics.ca.CAThread):
    '''A monitor dispatcher which works with pyepics

    The monitor dispatcher works around having callbacks from libca threads.
    Using epics CA calls (caget, caput, etc.) from those callbacks is not
    possible without this dispatcher workaround.

    ... note::

       Without `all_contexts` set, only the callbacks that are run with
       the same context as the the main thread are affected.

    ... note::

       Ensure that you call epics.ca.use_initial_context() at startup in
       the main thread

    Parameters
    ----------
    all_contexts : bool, optional
        re-route _all_ callbacks from _any_ context to the dispatcher callback
        thread
    timeout : float, optional
    callback_logger : logging.Logger, optional
        A logger to notify about failed callbacks

    Attributes
    ----------
    main_context : ctypes long
        The main CA context
    callback_logger : logging.Logger
        A logger to notify about failed callbacks
    queue : Queue
        The event queue
    '''

    def __init__(self, all_contexts=False, timeout=0.1,
                 callback_logger=None):
        epics.ca.CAThread.__init__(self, name='monitor_dispatcher')

        self.daemon = True
        self.queue = queue.Queue()

        # The dispatcher thread will stop if this event is set
        self._stop_event = threading.Event()
        self.main_context = epics.ca.current_context()
        self.callback_logger = callback_logger

        self._all_contexts = bool(all_contexts)
        self._timeout = timeout

        self.start()

    def run(self):
        '''The dispatcher itself'''
        self._setup_pyepics(True)

        while not self._stop_event.is_set():
            try:
                callback, args, kwargs = self.queue.get(True, self._timeout)
            except queue.Empty:
                pass
            else:
                try:
                    callback(*args, **kwargs)
                except Exception as ex:
                    if self.callback_logger is not None:
                        self.callback_logger.error(ex, exc_info=ex)

        self._setup_pyepics(False)
        epics.ca.detach_context()

    def stop(self):
        '''Stop the dispatcher thread and re-enable normal callbacks'''
        self._stop_event.set()

    def _setup_pyepics(self, enable):
        # Re-route monitor events to our new handler
        if enable:
            fcn = self._monitor_event
        else:
            fcn = epics.ca._onMonitorEvent

        epics.ca._CB_EVENT = (
            ctypes.CFUNCTYPE(None, epics.dbr.event_handler_args)(fcn))

    def _monitor_event(self, args):
        if (self._all_contexts or
                self.main_context == epics.ca.current_context()):
            if callable(args.usr):
                if (not hasattr(args.usr, '_disp_tag') or
                        args.usr._disp_tag is not self):
                    args.usr = lambda orig_cb=args.usr, **kwargs: \
                        self.queue.put((orig_cb, [], kwargs))
                    args.usr._disp_tag = self

        return epics.ca._onMonitorEvent(args)


def setup(logger):
    '''Setup ophyd for use

    Must be called once per session using ophyd
    '''
    try:
        epics.ca.find_libca()
        # if we can not find libca, then we clearly are not
        # going to be using CA threads so no need to install
        # the trampoline
    except epics.ca.ChannelAccessException:
        return
    # It's important to use the same context in the callback dispatcher
    # as the main thread, otherwise not-so-savvy users will be very
    # confused
    global _dispatcher

    if _dispatcher is not None:
        logger.debug('ophyd already setup')
        return

    def _cleanup():
        '''Clean up the ophyd session'''
        global _dispatcher
        if _dispatcher is None:
            return

        logger.debug('Performing ophyd cleanup')
        if _dispatcher.is_alive():
            logger.debug('Joining the dispatcher thread')
            _dispatcher.stop()
            _dispatcher.join()

        _dispatcher = None

        logger.debug('Finalizing libca')
        epics.ca.finalize_libca()

    epics.ca.use_initial_context()

    logger.debug('Installing monitor dispatcher')
    _dispatcher = MonitorDispatcher()
    atexit.register(_cleanup)
    return _dispatcher


def get_pv_form(version):
    '''Get the PV form that should be used for pyepics

    Due to a bug in certain versions of PyEpics, form='time' cannot be used
    with some large arrays.

    native: gives time.time() timestamps from this machine
    time: gives timestamps from the PVs themselves

    Returns
    -------
    {'native', 'time'}
    '''
    def _fix_git_versioning(in_str):
        return in_str.replace('-g', '+g')

    def _naive_parse_version(version):
        try:
            version = version.lower()

            # Strip off the release-candidate version number (best-effort)
            if 'rc' in version:
                version = version[:version.index('rc')]

            version_tuple = tuple(int(v) for v in version.split('.'))
        except Exception:
            return None

        return version_tuple

    try:
        from pkg_resources import parse_version
    except ImportError:
        parse_version = _naive_parse_version

    version = parse_version(_fix_git_versioning(version))

    if version is None:
        warnings.warn('Unrecognized PyEpics version; using local timestamps',
                      ImportWarning)
        return 'native'

    elif version <= parse_version('3.2.3'):
        warnings.warn('PyEpics versions <= 3.2.3 will use local timestamps (version: %s)' %
                      epics.__version__,
                      ImportWarning)
        return 'native'
    else:
        return 'time'


pv_form = get_pv_form(epics.__version__)
